/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */
package borland.jbcl.control.applet;

import java.util.Hashtable;
import java.util.Vector;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.PushbackInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.security.*;

import borland.jbcl.util.Diagnostic;

import sun.misc.MessageUtils;

/**
 * This class defines an applet class loader.
 *
 * @version     1.55, 01/21/97
 */
public class AppletClassLoader extends ClassLoader {

    /* Are we debugging? */
    private static boolean debug = false;

    /* Archives we keep track of for the resource API */
    Vector archives = new Vector();

    protected boolean verboseClassLoading = false;

    /* The codebase for this class loader. This is the basis for many
       security decisions. */
    URL base;

    /* The AppletResourceLoader from where to get resources and archives */
    AppletResourceLoader resources;

    /* The threadgroup that the applets loaded by this classloader live
       in. In the sun.* implementation of applets, the security manager's
       (AppletSecurity) getThreadGroup returns the thread group of the
       first applet on the stack, which is the applet's thread group. */
    AppletThreadGroup threadGroup;

    /** A hash table of AppletClassEntry's that define .class files in their
     *  raw form.  We <B>do not</B> define these classes immediately
     *  in defineClassFromBytes(), instead the byte arrays are kept
     *  around and classes are defined lazily on demand, and the byte
     *  array is removed from the Hashtable at that time.
     *  <P>
     *  By defining classes lazily, we prevent the case where a
     *  derived class was found in a jar before its super class also
     *  in the jar.  The derived class cannot be defined before it's
     *  super class has even been loaded from the archive.
     *
     *  classes are hashed by fully qualified name - "my.package.MyApplet"
     */

    protected Hashtable rawClasses = new Hashtable();

    /**
     * Load applets from a base URL.
     */
    protected AppletClassLoader(URL base) {
        this.base = base;
        this.resources = new AppletResourceLoader(this);
    }

    /**
     * Load applets from a base URL.
     */
    protected AppletClassLoader(URL base, AppletResourceLoader resources) {
        this.base = base;
        this.resources = resources;
    }

    /**
     * Return the code base for this class loader.
     */
    protected URL getCodeBase() {
        return base;
    }

    /**
     * Load a class from a URL.
     */
    private Class loadClass(String name, URL url) throws IOException {
        InputStream in = null;
        try {
            URLConnection c = url.openConnection();
            c.setAllowUserInteraction(false);
            in = c.getInputStream();

            int len = c.getContentLength();
            byte data[] = new byte[(len == -1) ? 4096 : len];
            int total = 0, n;

            Thread thread = Thread.currentThread();

            while (((n = in.read(data, total, data.length - total)) >= 0)
                   && !thread.isInterrupted()) {
                if ((total += n) == data.length) {
                    if (len < 0) {
                        byte newdata[] = new byte[total * 2];
                        System.arraycopy(data, 0, newdata, 0, total);
                        data = newdata;
                    } else {
                        break;
                    }
                }
            }
            if (thread.isInterrupted()) {
                throw new AppletIOException("loadclass.throw.interrupted", url);  //NORES

            }

            return defineClass(name, data, 0, total);

        } catch (IOException e) {
            Diagnostic.printStackTrace(e);
            throw e;
        } catch (Throwable e) {
            Diagnostic.printStackTrace(e);
            throw new AppletIOException("loadclass.throw.notloaded", url);  //NORES
        } finally {
            if (in != null) {
                in.close();
            }
        }
    }

    /**
     * Load a class from this class loader.
     */
    public Class loadClass(String name) throws ClassNotFoundException {
        return loadClass(name, true);
    }

    /**
     * Load and resolve a class.
     */
    protected Class loadClass(String name, boolean resolve)
                    throws ClassNotFoundException {

        /* We check to see if we've already loaded it (i.e. in the cache) */
        Class cl = findLoadedClass(name);

        /* If not, we check to see if someone is allowed to _define_ a class
         * in a package */
        if (cl == null) {
            SecurityManager security = System.getSecurityManager();
            if (security != null) {
                int i = name.lastIndexOf('.');
                if (i >= 0) {
                    /* REMIND: is this the check we want to make? */
                    security.checkPackageAccess(name.substring(0, i));
                }
            }

            /* This is a sensitive piece of code. It used to catch
             * Throwable, which may have caused findSystemClass to
             * unduly fail (i.e. on "String") as an applet called
             * stop(Throwable t) on the thread doing the loading. */
            try {
                return findSystemClass(name);
            } catch (ClassNotFoundException e) { }

            cl = findClass(name);
        }
        if (cl == null) {
            throw new ClassNotFoundException(name);
        }
        if (resolve) {
            resolveClass(cl);
        }
        return cl;
    }

    /**
     * Load a class from a specific file, unless the class has already been
     * loaded.  This method is synchronized so that two applets do not load
     * the same file simultaneously.
     */
    private synchronized Class loadCode(String root, String extension,
                                        String className)
        throws ClassNotFoundException, IOException {

        /* See if the class was already loaded */
        Class ourClass = findLoadedClass(className);
        if (ourClass != null) {
            return ourClass;
        }

        /* see if it's a raw class */
        AppletClassEntry en = (AppletClassEntry) rawClasses.get(className);
        if (en != null) {
            try {
                Class c = super.defineClass(className, en.classBuf, en.start, en.len);
                if (en.ids != null) {
                    setSigners(c, en.ids);
                }
                return c;
            } finally {
                rawClasses.remove(className);
            }
        }

        /*
         * It's not loaded, check to see if we are allowed to access and
         * define it.
         */

        SecurityManager security = System.getSecurityManager();
        if (security != null) {
            int packageIndex = className.lastIndexOf('.');
            if (packageIndex >= 0) {
                String packageName = className.substring(0, packageIndex);
                security.checkPackageAccess(packageName);
            }
        }
        /* Try to load the class using the system class loader */
        try {
            return findSystemClass(className);
        } catch (ClassNotFoundException ex) {
            /* Oh well, try something else */
        }
        try {
            /* Try to load the class from the URL */
            URL url = new URL(base, root + extension);
            verbose(amh.getMessage("loadcode.verbose", url.toString(),
                                   className));
            ourClass = loadClass(className, url);
        } catch (FileNotFoundException e) {
            errorMsg("filenotfound", className);
        } catch (ClassFormatError e) {
            errorMsg("fileformat", className);
        } catch (IOException e) {
            throw e;
        } catch (Exception e) {
            errorMsg("fileexception", e.toString(), className);
        } catch (ThreadDeath e) {
            errorMsg("filedeath", className);
            throw e;
        } catch (Error e) {
            errorMsg("fileerror", e.toString(), className);
        }
        if (ourClass != null) {
            /* Resolve the class */
            resolveClass(ourClass);
        }
        return ourClass;
    }

    /**
     * Load and resolve the file specified by the code= attribute.
     * This is quite like loadClass(String), except the parameter is
     * a file name, not a class name.
     * This method is package-private since it should only be called from
     * sunw.applet.AppletPanel.
     */
    Class loadCode(String filename) throws ClassNotFoundException {
        debug("AppletClassLoader(" + base.toString() +
                ").loadCode(" + filename + ")");
        /*
         * there are essentially 4 variants of file/class names
         * that might be used to reference a class:
         *
         *    name specified    class name and physical file name
         *
         * 1) foo/bar.class ->  foo.bar from foo/bar.class
         *
         * 2) foo/bar.what ->   foo.bar from foo/bar.what
         *
         * 3) foo.bar.class ->  foo.bar from foo/bar.class
         *    foo.bar.java ->   foo.bar from foo/bar.class
         *
         * 4) foo.bar.what ->   foo.bar.what from foo/bar/what.class
         *              or ->   foo.bar from foo/bar.what
         *              or ->   foo.bar.what from foo/bar.what
         *
         * In the first three cases, the mapping is simple and
         * easily derived.  The first two cases are correct per
         * specification, and are essentially identical.  The
         * third and fourth cases are typically allowed by current
         * browsers, although in the fourth case, there is some
         * ambiguity as to the intent, thus causing us to attempt
         * open multiple files.
         *
         * Also, in cases 3 and 4, we substite the slashes for dots, etc.
         */
        String defaultExtension = ".class";
        String oldExtension = ".java";
        String root;
        String className;
        int extensionIndex = filename.lastIndexOf('.');
        String extension = "";

        if (extensionIndex != -1) {
            extension = filename.substring(extensionIndex);
        }

        if (extensionIndex != -1 && filename.indexOf('/') != -1) {

            // handle cases 1 and 2
            root = filename.substring(0, extensionIndex);
            className = root.replace('/', '.');

            try {
                return loadCode(root, extension, className);
            } catch (IOException e) {
                throw new ClassNotFoundException(className);
            }

        // Replace older uses of ".java" as an extension with ".class"
        } else if (extension.equals(oldExtension) ||
            extension.equals(defaultExtension)) {
            // handle case 3
            root = filename.substring(0, extensionIndex);
            root = root.replace('.', '/');

            extension = defaultExtension;
            className = root.replace('/', '.');

            try {
                return loadCode(root, extension, className);
            } catch (IOException e) {
                throw new ClassNotFoundException(className);
            }
        }

        // case 4

        /*
         * Now, get the class name of the class in the file.
         * This looks like you could be spoofed, but the class loader checks
         * that this name matches the name of the class in the file.
         */
        root = filename.replace('.', '/');
        extension = defaultExtension;
        className = root.replace('/', '.');

        try {
            return loadCode(root, extension, className);
        } catch (java.io.IOException e) {
            // do nothing, and try next alternate
        }catch (ClassNotFoundException e) {
            // do nothing, and try next alternate
        }

        if (extensionIndex < 0) {
            extensionIndex = filename.length();
        }

        root = filename.substring(0, extensionIndex);
        root = root.replace('.', '/');
        extension = filename.substring(extensionIndex);
        className = root.replace('/', '.');

        try {
            return loadCode(root, extension, className);
        } catch (java.io.IOException e) {
            // do nothing, and try next alternate
        }catch (ClassNotFoundException e) {
            // do nothing, and try next alternate
        }

        root = filename.substring(0, extensionIndex);
        root = root.replace('.', '/');
        extension = filename.substring(extensionIndex);
        className = filename.replace('/', '.');

        try {
            return loadCode(root, extension, className);
        } catch (IOException e) {
            throw new ClassNotFoundException(className);
        }
    }


    /*
     * This method exists so we can read a class from a Jar stream and
     * get it defined.  It should have package-only access.
     *
     * If it's loaded already, we're happy. Otherwise we check for
     * definition priviledges- if it passes, we check to see if there
     * is a system class with that name. If not, we define it!
     */
    AppletClassEntry defineClassFromBytes(String name, byte[] classbuf,
                              int start, int len) {

        SecurityManager security = System.getSecurityManager();
        String packageName = "";
        int i = name.lastIndexOf('.');
        if (i >= 0) {
            packageName = name.substring(0, i);
        }
        if (security != null) {
            security.checkPackageDefinition(packageName);
        }

        AppletClassEntry ent = new AppletClassEntry(classbuf, start, len, name, null);
        rawClasses.put(ent, ent);
        rawClasses.put(name, ent);

        return ent;
        // REMIND pass the name
        //      return super.defineClass(classbuf, start, len);


    }

    void errorMsg(String key, String arg) {
        System.err.println(amh.getMessage(key, arg));

    }

    void errorMsg(String key, String arg1, String arg2) {
        System.err.println(amh.getMessage(key, arg1, arg2));
    }

    /**
     * This method finds a class. The returned class
     * may be unresolved. This method has to be synchronized
     * to avoid two threads loading the same class at the same time.
     * Must be called with the actual class name.
     */
    private synchronized Class findClass(String name) {
        Class cl = findLoadedClass(name);
        if (cl != null) {
            return cl;
        }

        /* see if it's a raw class */
        AppletClassEntry en = (AppletClassEntry) rawClasses.get(name);
        if (en != null) {
            try {
                Class c = super.defineClass(name, en.classBuf, en.start, en.len);
                if (en.ids != null) {
                    setSigners(c, en.ids);
                }
                return c;
            } finally {
                rawClasses.remove(name);
            }
        }


        verbose(amh.getMessage("findclass.verbose.findclass", Thread.currentThread().getName(), name));


        SecurityManager security = System.getSecurityManager();
        if (security != null) {
            int i = name.lastIndexOf('.');
            if (i >= 0) {
                security.checkPackageDefinition(name.substring(0, i));
            }
        }

        String cname = name.replace('.', '/') + ".class";

        try {
            URL url = new URL(base, cname);
            verbose(amh.getMessage("findclass.verbose.openstream", url.toString(),
                                   name));
            cl = loadClass(name, url);
        } catch (FileNotFoundException e) {
            errorMsg("filenotfound", name);
        } catch (ClassFormatError e) {
            errorMsg("fileformat", name);
        } catch (IOException e) {
            errorMsg("fileioexception", name);
        } catch (Exception e) {
            errorMsg("fileexception", e.toString(), name);
        } catch (ThreadDeath e) {
            errorMsg("filedeath", name);
            throw e;
        } catch (Error e) {
            errorMsg("fileerror", e.toString(), name);
        }
        return cl;
    }

    /**
     * Resource locator methods
     */

    /**
     * The resource loader that this class loader relies on
     */
    public AppletResourceLoader getResourceLoader() {
        return resources;
    }

    private boolean resourceExists(URL url) {
        // Check if the resource exists.
        // It almost works to just try to do an openConnection() but
        // HttpURLConnection will return true on HTTP_BAD_REQUEST
        // when the requested name ends in ".html", ".htm", and ".txt"
        // and we want to be able to handle these
        //
        // Also, cannot just open a connection for things like FileURLConnection,
        // because they suceed when connecting to a non-existant file.
        // So, in those cases we open and close an input stream.
        boolean ok = true;
        try {
            URLConnection conn = url.openConnection();
            if (conn instanceof java.net.HttpURLConnection) {
                java.net.HttpURLConnection hconn =
                    (java.net.HttpURLConnection) conn;
                int code = hconn.getResponseCode();
                if (code == java.net.HttpURLConnection.HTTP_OK) {
                    return true;
                } if (code >= java.net.HttpURLConnection.HTTP_BAD_REQUEST) {
                    return false;
                }
            } else {
                // our best guess for the other cases
                InputStream is = url.openStream();
                is.close();
            }
        } catch (Exception ex) {
            ok = false;
        }
        return ok;
    }

    public java.net.URL getResource(String name) {
        verbose(amh.getMessage("getresource.verbose.forname", name));
        java.net.URL back = getSystemResource(name);
        if (back != null) {
            verbose(amh.getMessage("getresource.verbose.found", name));
            return back;
        } else {
            // REMIND - we should deal only in URLs
            String urlName = resources.getResourceAsName(name);
            try {
                back = new java.net.URL(urlName);
                if (! resourceExists(back)) {
                    back = null;
                }
            } catch (Exception ex) {
                back = null;
            }
        }
        return back;
    }

    public InputStream getResourceAsStream(String name) {
        // REMIND - This should be built on getResource()
        verbose("AppletClassLoader.getResourceAsStream for name: "+name);
        // will locate the resource in this
        InputStream back = getSystemResourceAsStream(name);
        if (back != null) {
            verbose(amh.getMessage("getresourceasstream.verbose", name));
            return back;
        }
        return resources.getResourceAsStream(name);
    }

    /**
     * This method returns
     */
    protected synchronized ThreadGroup getThreadGroup() {
        if (threadGroup == null || threadGroup.isDestroyed()) {
            threadGroup = new AppletThreadGroup(base + "-threadGroup");
            threadGroup.setDaemon(true);
            threadGroup.setMaxPriority(Thread.NORM_PRIORITY - 1);
        }
        return threadGroup;
    }

    private void debug(String s) {
        if (debug)
            System.err.println(s);
    }

    private void verbose(String s) {
        if (verboseClassLoading) {
            System.err.println(s);
        }
    }
    private static AppletMessageHandler amh = new AppletMessageHandler("appletclassloader");

}
