/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */
///--------------------------------------------------------------------------------------------------
// Copyright (c) 1996-1998 Borland International, Inc. All Rights Reserved.
//--------------------------------------------------------------------------------------------------

package borland.jbcl.io;

import java.io.*;
import borland.jbcl.util.Diagnostic;
import borland.jbcl.util.Hex;

public class AsciiOutputStream extends SimpleCharOutputStream {

  private byte[]        buffer;
  private int           count;
  private int           bufferLength;
  private int           backCount;
  private OutputStream  out;

  /**
   * Creates a new buffered stream with a default
   * buffer size.
   * @param out       the output stream
   */
  public AsciiOutputStream(OutputStream out) {
    this(out, 2048);
  }

  /**
   * Creates a new buffered stream with the specified
   * buffer size.
   * @param out       the output stream
   * @param bufferLength      the buffer size
   */
  public AsciiOutputStream(OutputStream out, int bufferLength) {
    this.out          = out;
    this.count        = -1;
    this.bufferLength = bufferLength;
    buffer            = new byte[bufferLength];
  }

  /**
   * Writes a char. This method will block until the byte is actually
   * written.
   * @param b the byte to be written
   * @exception IOException If an I/O error has occurred.
   *
   * Ascii encoding scheme. compiler will not compile with '\' in front of u, so I
   * use <bs> instead.
   *
   * Single '<bs>' ignored unless it is for '<bs>u' (which is interpreted)
   * <bs>       ->  <bs>
   * <bs>n      ->  <bs>n
   * <bs>u      ->  <bs><bs>u
   * Consecutive <bs> escaped by doubling thier number (making thier count even).
   * <bs><bs>      -> <bs><bs><bs><bs>
   * Outputing real unicode character, make sure that odd number of '<bs>' precede.
   * 0x7ff  ->  <bs>u07ff
   * <bs>0x7ff  -> <bs><bs><bs>u07ff
   *
   * 0-0xFF characters are not encoded.  This allows all Western European characters
   * to go out as is in the hopes that some other existing products may be able
   * to read this file format.  For 0xFF00-0x0100 <bs>u encoding is used.  For
   * this range, I doubt any existing products (other than ours) will be to
   * expect this format.
   */
  public void write(int ch) throws IOException {
    if ((ch & 0xFF00) != 0) {
      writeByte('\\');
      if (backCount > 0) {
        writeByte('\\');
        backCount = 0;
      }
      writeByte('u');
      writeByte(Hex.chars[(ch >> 12) & 0xF]);
      writeByte(Hex.chars[(ch >>  8) & 0xF]);
      writeByte(Hex.chars[(ch >>  4) & 0xF]);
      writeByte(Hex.chars[ch & 0xF]);
    }
    else if (ch == '\\') {
      writeByte(ch);
      if (++backCount > 1) {
        writeByte(ch);
        writeByte(ch);
        backCount = 0;
      }
    }
    else if (backCount > 0) {
      if (ch == 'u') {
        writeByte('\\');
      }
      writeByte(ch);
      backCount = 0;
    }
    else if (++count < bufferLength) {
      buffer[count] = (byte)ch;
    }
    else {
      flush();
      writeByte(ch);
    }
  }

  private void writeByte(int ch) throws IOException {
    if (++count < bufferLength)
      buffer[count] = (byte)ch;
    else {
      flush();
      buffer[++count] = (byte)ch;
    }
  }

  public void flush() throws IOException {
    if (count > 0) {
      out.write(buffer, 0, count);
      out.flush();
      count = -1;
    }
  }

  public void close() throws IOException {
    //Diagnostic.println("buffer.length "+buffer.length+" count "+count);
    flush();
    out.close();
  }
}
