/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */
///--------------------------------------------------------------------------------------------------
// Copyright (c) 1996-1998 Borland International, Inc. All Rights Reserved.
//--------------------------------------------------------------------------------------------------

package borland.jbcl.io;

import java.io.*;
import sun.io.*;
import borland.jbcl.util.Hex;

/**
 *  An implementation of a CharToByteConverter that includes conversion
 *  to Unicode escapes. <p>
 *
 *  All the conversions in sun.io, when they encounter a character that
 *  is not in the target character set, either throw an 
 *  UnknownCharacterException, or use a substitution character.
 *  This class uses an underlying standard conversion provided in
 *  sun.io, but inserts Unicode escapes instead for unknown characters.
 *  This is useful for writing to .java source files.
 *  Since this is a CharToByteConverter, it can be passed to text output
 *  streams (the Writer classes), which are the new recommended way
 *  to output text files. <p>
 *
 *  Custom CharToByteConverters (outside of sun.io), such as this one, 
 *  cannot be created using CharToByteConverter.getConverter(). 
 *  Use one of the constructors instead, e.g. 
 *  <pre>
 *    conv = new CharToByteJava("SJIS"). 
 *    OutputStreamWriter = new OutputStreamWriter(os, conv);
 *  </pre>
 */

public class CharToByteJava extends CharToByteConverter {
  
  /**
   * Creates a converter from an existing converter (presumably
   * from sun.io, created by CharToByteConverter.getConverter()).
   * @param conv the existing converter
   */
  public CharToByteJava(CharToByteConverter conv) {
    this.conv = conv;
    conv.reset();
  }

  /**
   * Creates a converter with the named encoding.
   * @param conv  the encoding name, as supported by 
   *              CharToByteConverter.getConverter.
   */
  public CharToByteJava(String encoding) throws UnsupportedEncodingException {
    conv = CharToByteConverter.getConverter(encoding);
  }

  /**
   * Creates a converter with the default encoding.
   * The default encoding is determined with the following priority.
   * <pre>
   * (1) property "latte.compiler.file.encoding.write"
   * (2) property "latte.compiler.file.encoding"
   * (3) property "file.encoding"
   * (4) the default encoding returned by CharToByteConverter.getDefault().
   * </pre>
   */
  public CharToByteJava()  {
    String name = System.getProperty("latte.compiler.file.encoding.write");  //NORES
    if (name == null) {
      name = System.getProperty("latte.compiler.file.encoding");  //NORES
    }
    if (name == null) {
      name = System.getProperty("file.encoding");  //NORES
    }
    if (name != null) {
      try {
        conv = CharToByteConverter.getConverter(name);
        return;
      }
      catch (UnsupportedEncodingException e) {
        // will exit with CharToByteConverter.getDefault() below
      }
    }
    conv = CharToByteConverter.getDefault();
  }

  /**
   * Returns the name for the encoding.
   * This is always formatted as the name of the underlying
   * sun.io converter, followed by ":Java". E.g. a converter
   * constructed with new CharToByteJava("SJIS") will return
   * an encoding name of "SJIS:Java".
   */
  public String getCharacterEncoding()  { 
    return conv.getCharacterEncoding() + ":Java";  //NORES
  }

  public int convert(char[] input, int inStart, int inEnd, byte[] output, int outStart, int outEnd)
      throws MalformedInputException, ConversionBufferFullException
  {
    charOff = inStart;
    byteOff = outStart;
    conv.setSubstitutionMode(false);
    while (charOff < inEnd) {
      try {
        byteOff += conv.convert(input, charOff, inEnd, output, byteOff, outEnd);
        charOff = conv.nextCharIndex();
      }
      catch (UnknownCharacterException e) {
        byteOff = conv.nextByteIndex();
        charOff = conv.nextCharIndex();
        if (outEnd - byteOff < 6) {
    	    throw new ConversionBufferFullException();
        }
        char ch = input[charOff++];
        output[byteOff++] = '\\';
        output[byteOff++] = 'u';
        output[byteOff++] = Hex.bytes[ch>>12];
        output[byteOff++] = Hex.bytes[(ch>>8) & 0xf];
        output[byteOff++] = Hex.bytes[(ch>>4) & 0xf];
        output[byteOff++] = Hex.bytes[ch & 0xf];
      }
      catch (ConversionBufferFullException e) {
        byteOff = conv.nextByteIndex();
        charOff = conv.nextCharIndex();
  	    throw new ConversionBufferFullException();
      }
    }
    int res = byteOff - outStart;
    return byteOff-outStart;
  }

  public int flush(byte[] output, int outStart, int outEnd)
      throws MalformedInputException, ConversionBufferFullException
  {
    return conv.flush(output, outStart, outEnd);
  }

  public void reset()  {
    conv.reset();
  }

  public int getMaxBytesPerChar()  {
    return 6;
  }

  /**
   * Returns the index of the character just past
   * the last character successfully converted by the previous call
   * to convert.
   */
  public int nextCharIndex() {
      return charOff;
  }

  /**
   * Returns the index of the byte just past the last byte written by
   * the previous call to convert.
   */
  public int nextByteIndex() {
      return byteOff;
  }

  /**
   * Returns true if the given character can be converted to the
   * target character encoding.
   * @return true if given character is translatable, false otherwise.
   * @param c character to test
   */
  public boolean canConvert(char c) {
    return true;
  }

  private CharToByteConverter conv;
}
