/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */
///--------------------------------------------------------------------------------------------------
// Copyright (c) 1996-1998 Borland International, Inc. All Rights Reserved.
//--------------------------------------------------------------------------------------------------
package borland.jbcl.io;

import java.io.*;
import sun.io.CharToByteConverter;
import sun.io.ConversionBufferFullException;


/**
 * Write characters to an output stream, translating characters into bytes
 * according to a specified character encoding.  Characters that cannot
 * be represented in the specified encoding will be represented by
 * Unicode escapes.
 */
public class JavaOutputStreamWriter extends OutputStreamWriter {

    private CharToByteJava ctb;
    private OutputStream out;

    private static final int defaultByteBufferSize = 8192;
    private byte bb[];
    private int nextByte = 0;
    private int nBytes = 0;

    /**
     * Create an JavaOutputStreamWriter that uses the named character encoding.
     *
     * @param  out  An OutputStream
     * @param  enc  Name of the encoding to be used
     *
     * @exception  UnsupportedEncodingException
     *             If the named encoding is not supported
     */
    public JavaOutputStreamWriter(OutputStream out, String enc)
        throws UnsupportedEncodingException
    {
        this(out, new CharToByteJava(enc));
    }

    /**
     * Create an JavaOutputStreamWriter that uses the default character encoding.
     *
     * @param  out  An OutputStream
     */
    public JavaOutputStreamWriter(OutputStream out) {
        this(out, new CharToByteJava());
    }

    /**
     * Create an JavaOutputStreamWriter that uses the specified character-to-byte
     * converter.  The converter is assumed to have been reset.
     *
     * @param  out  An OutputStream
     * @param  ctb  A CharToByteConverter
     */
    private JavaOutputStreamWriter(OutputStream out, CharToByteConverter ctb) {
        super(out);
        this.out = out;
        this.ctb = new CharToByteJava(ctb);
        bb = new byte[defaultByteBufferSize];
        nBytes = defaultByteBufferSize;
    }

    /**
     * Return the name of the encoding being used by this stream.  May return
     * null if the stream has been closed.
     */
    public String getEncoding() {
        synchronized (lock) {
            if (ctb != null)
                return ctb.getCharacterEncoding();
            else
                return null;
        }
    }

    /** Check to make sure that the stream has not been closed */
    private void ensureOpen() throws IOException {
        if (out == null)
            throw new IOException("Stream closed");
    }

    /**
     * Write a single character.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void write(int c) throws IOException {
        write(new char[] {(char)c}, 0, 1);
    }

    /**
     * Write a portion of an array of characters.
     *
     * @param  cbuf  Buffer of characters
     * @param  off   Offset from which to start writing characters
     * @param  len   Number of characters to write
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void write(char cbuf[], int off, int len) throws IOException {
        synchronized (lock) {
            ensureOpen();

            int ci = off, end = off + len;
            while (ci < end) {
                boolean bufferFull = false;

                try {
                    nextByte += ctb.convert(cbuf, ci, end,
                                            bb, nextByte, nBytes);
                    ci = end;
                }
                catch (ConversionBufferFullException x) {
                    int nci = ctb.nextCharIndex();
                    if (nci == ci) {
                        /* Buffer doesn't even hold one character */
                        throw new CharConversionException("Output buffer too small");
                    }
                    ci = nci;
                    bufferFull = true;
                    nextByte = ctb.nextByteIndex();
                }

                if ((nextByte >= nBytes) || bufferFull) {
                    out.write(bb, 0, nextByte);
                    nextByte = 0;
                }
            }
        }
    }

    /**
     * Write a portion of a string.
     *
     * @param  str  A String
     * @param  off  Offset from which to start writing characters
     * @param  len  Number of characters to write
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void write(String str, int off, int len) throws IOException {
        char cbuf[] = new char[len];
        str.getChars(off, off + len, cbuf, 0);
        write(cbuf, 0, len);
    }

    /**
     * Flush the output buffer to the underlying byte stream, without flushing
     * the byte stream itself.  This method is non-private only so that it may
     * be invoked by PrintStream.
     */
    void flushBuffer() throws IOException {
        synchronized (lock) {
            ensureOpen();

            for (;;) {
                try {
                    nextByte += ctb.flush(bb, nextByte, nBytes);
                }
                catch (ConversionBufferFullException x) {
                    nextByte = ctb.nextByteIndex();
                }
                if (nextByte == 0)
                    break;
                if (nextByte > 0) {
                    out.write(bb, 0, nextByte);
                    nextByte = 0;
                }
            }
        }
    }

    /**
     * Flush the stream.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void flush() throws IOException {
        synchronized (lock) {
            flushBuffer();
            out.flush();
        }
    }

    /**
     * Close the stream.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void close() throws IOException {
        synchronized (lock) {
            if (out == null)
                return;
            flush();
            out.close();
            out = null;
            bb = null;
            ctb = null;
        }
    }

}
