/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */

package borland.jbcl.io;

import java.io.*;
import java.util.*;
import java.util.StringTokenizer;

import borland.jbcl.util.*;

public class LocalFileSystem implements FileSystem
{
  // To make this abstraction complete need a corresponding File interface
  // and supporting object(s).  Just deal with plain File for now.
  //
  // To bring complete FileMan functionality here, need native functions for
  // delete directory (only have delete file), and setLastModified().
  //
  public String createFilePath(String path, String file) {
    return path + File.separator + file;
  }

  public void copyFile(String src, String dest) throws IOException {
    copyFile(src, dest, null);
  }

  public void copyFile(String src, String dest, String header) throws IOException {
    // Use unbuffered streams, because we're going to use a large buffer
    // for this sequential io.
    FileInputStream  input   = new FileInputStream(src);
    FileOutputStream output  = new FileOutputStream(dest);

    if (header != null) {
      int  headerLength  = header.length();
      byte[] headerBytes = new byte[headerLength];
      header.getBytes(0, headerLength, headerBytes, 0);
      output.write(headerBytes, 0, headerLength);
    }

    int bytesRead;
    byte[] buffer = new byte[32*1024];
    while ((bytesRead = input.read(buffer, 0, buffer.length)) > 0)
      output.write(buffer, 0, bytesRead);

    input.close();
    output.close();
  }

  public boolean compareBytes(byte[] buf1, byte[] buf2, int count) {
    for (int i = 0; i < count; i++)
      if (buf1[i] != buf2[i])
        return false;
    return true;
  }

  public boolean compareFiles(String path1, String path2) throws IOException {
    long total = newFile(path1).length();
    if (total != newFile(path2).length())
      return false;

    FileInputStream input1 = new FileInputStream(path1);
    FileInputStream input2 = new FileInputStream(path2);

    byte[] buffer1 = new byte[32*1024];
    byte[] buffer2 = new byte[32*1024];

    long totalRead = 0;
    int  bytesRead;
    while ((bytesRead = input1.read(buffer1, 0, buffer1.length)) > 0) {
      if (input2.read(buffer2, 0, buffer2.length) != bytesRead)
        return false;
      if (!compareBytes(buffer1, buffer2, bytesRead))
        return false;
      totalRead += bytesRead;
    }
    input1.close();
    input2.close();
    return total == totalRead;
  }

  public void deleteDirectoryTree(String dirPath) throws Exception {
    Diagnostic.fail();
    /*
    Don't have directory deletion without native methods.

    File  dirFile = newFile(dirPath);
    DirectoryDeleter  deleter = new DirectoryDeleter(dirFile);
    traverse(deleter, dirFile);
    deleteDirectory(dirPath);
    */
  }

  public boolean traverse(TraverseAction act, File dir) throws Exception {
    if (!dir.isAbsolute())
      return false;

    String[] fileList = dir.list();
    if (fileList == null)
      return false;

    for (int i = 0; i < fileList.length; i++) {
      File curFile = newFile(createFilePath(dir.getAbsolutePath(), fileList[i]));
      if (curFile.isFile()) {
        if (act.onFile(curFile) == act.ABORT_TRAVERSE)
          return false;
      }
      else if (curFile.isDirectory()) {
        if (act.onDirEntry(curFile) == act.DIVE_TRAVERSE) {
          if (!traverse(act, newFile(curFile.getAbsolutePath()))
              || act.onDirExit(curFile, fileList) != act.CONTINUE_TRAVERSE)
            return false;
        }
      }
    }
    return true;
  }

  public Properties loadProperties(String path) throws IOException {
    Properties properties  = new Properties();
    BufferedInputFile stream  = new BufferedInputFile(path);
    properties.load(stream);
    stream.close();
    return properties;
  }

  public boolean makeDirs(String path) {
    File dir = newFile(path);
    dir.mkdirs();
    return dir.isDirectory();
  }

  public boolean makeDir(String path) {
    File dir = newFile(path);
    dir.mkdir();
    return dir.isDirectory();
  }

  public String addDir(String path, String dir) {
    File file = newFile(path);
    return file.getParent() + File.separator + dir + File.separator + file.getName();
  }

  // This is in here because the list() method of the File class returns a String[]
  public int listContains(String[] baseList, String name) {
    for (int i = 0; i < baseList.length; i++) {
      // Have to use IgnoreCase for netware server with long filename support.
      if (baseList[i].equalsIgnoreCase(name))
        return i;
    }
    return -1;
  }

  public File newFile(String path) {
    return new File(path);
  }

  public File getClassesFile(String fileName) {
    String path = System.getProperty("java.class.path");  //NORES
    StringTokenizer tokenizer = new StringTokenizer(path, File.pathSeparator);
    while (tokenizer.hasMoreTokens()) {
      path = tokenizer.nextToken();
      File file = new File(path + File.separator + fileName);
      if (file.exists())
        return file;
    }
    return null;
  }

  public File getClassesFile(Object object) {
    String name = object.getClass().getName().replace('.', File.separatorChar);
    name  = name + ".class";
    return getClassesFile(name);
  }

  public String getClassesDirectory(Object object) {
    return getClassesFile((Object)object).getParent();
  }
}
