/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */

package borland.jbcl.model;

import java.util.*;
import java.text.*;
import java.sql.*;

import borland.jbcl.util.*;

/**
 * This class extends the ItemFormatterclass through the use of
 * String patterns to control formatting and parsing.  Though other
 * implementations of Formatter are allowed, ItemFormatStr is the basic
 * one (and only one) built into the product at this time.
 *
 * Note that the current implementation supports only Variants and is
 * only a wrapper for VariantFormatStr.
 *
 * There are 4 different kinds of pattern strings which can be used.
 * Each is distinct, and the fields from one cannot be used with another.
 * The type used will be inferred from the Variant.type passed into the
 * constructor.
 * The three types are:
 *     ItemFormatter.NUMERIC  (double)
 *     ItemFormatter.DECIMAL  (BigDecimal)
 *     ItemFormatter.DATETIME
 *     ItemFormatter.TEXT
 *
 *
 * Following are detailed discussions of each of the characters allowed in
 * each of these patterns.
 *
 * For Numeric fields
 *
 *  Symbol  Meaning          Notes
 *    0    a digit
 *    #    a digit, zero shows as absent
 *    .    placeholder for decimal separator
 *    ,    placeholder for grouping delimiter.  Shows the interval to be used.
 *    ;    separates formats.                       There are two: positive and negative.
 *    %    divide by 100 and show as percentage
 *    X    any other characters can be used in
 *    -    a leading minus sign will appear for
 *    negative numbers
 *    ()    Alternatively, the entire expression
 *    can be enclosed in parentheses to show
 *    as negative (e.g. "($#,###.##)")
 *    {}    Can be used to bracket optional fields
 *    (e.g. "0000.{00}" will allow the decimal
 *    fraction to be omitted when editing
 *    ^    Sets the initial cursor position when editing
 *    \nnnn  Single character literal (e.g. \002, \0x0A, \u2030)
 *
 *  The numeric format mask actually consist of 2 semicolon separated masks.
 *  The first is required.  The second, if provided, will determine how negative
 *  numbers are formatted.  For example, "###.##;(###.##)" will format negative
 *  numbers in parentheses.
 *
 *
 *  For Date or TimeStamp fields:
 *
 *   Symbol  Meaning      Presentation  Notes
 *
 *    G    era designator    (Text)    AD/BC
 *    y    year       (Number)   Y,YY = 97, YYYY = 1997
 *    M    month in year    (Text & Number) M,MM = numeric, MMM = month abbrev, MMMM = full month
 *    d    day in month    (Number)  d,dd = 10, dddd = 0010
 *    h    hour in am/pm (1~12)  (Number)  h,hh = 12, hhhh = 0012
 *    H    hour in day (0~23)  (Number)  H,HH = 23, HHHH = 0023
 *    m    minute in hour    (Number)  m,mm = 59, mmmm = 0059
 *    s    second in     (Number)  s,ss = 59, ssss = 0059
 *    S    millisecond    (Number)  S = 9, SS = 99, SSS = 999
 *    E    day in week    (Text)    E,EE,EEE = Sun, EEEE = Sunday
 *    D    day in year    (Number)  D,DD = 364, DDDD = 0364
 *    F    day of week in month  (Number)   2 (2nd Wed in July)
 *    w    week in year    (Number)  27
 *    W    week in month    (Number)   2
 *    a    am/pm marker    (Text)     AM/PM
 *    k    hour in day (1~24)  (Number)  24
 *    K    hour in am/pm (0~11)  (Number)  0
 *    z    time zone    (Text)     z,zz,zzz = PDT, zzzz = Pacific Standard Time
 *    '    escape for text
 *    ''   single quote
 *    \nnnn  Single character literal (e.g. \002, \0x0A, \u2030)
 *    {}    Can be used to bracket optional fields
 *    (e.g. "0000.{00}" will allow the decimal
 *    fraction to be omitted when editing
 *    ^    Sets the initial cursor position when editing
 *
 * For string fields:
 * Symbol  Meaning
 *
 *    0    Digit (0 through 9, entry required; plus [+] and minus [-] signs not allowed).
 *    9    Digit or space (entry not required; plus and minus signs not allowed).
 *    #    Digit or space (entry not required; blank positions converted to spaces, plus and minus signs allowed).
 *    L    Letter (A through Z, entry required).
 *    l    Letter (A through Z, entry optional).
 *    ?    Letter (A through Z, entry optional).
 *    A    Letter or digit (entry required).
 *    a    Letter or digit (entry optional).
 *    C    Any character or a space (entry optional).
 *    c    Any character or a space (entry required).
 *    &    Any character or a space (entry required).
 *    <    Causes all characters that follow to be converted to lowercase.
 *    >    Causes all characters that follow to be converted to uppercase.
 *    !    Causes input mask to fill from right to left, rather than from left to right,
 *    when characters on the left side of the input mask are optional. You can include
 *    the exclamation point anywhere in the input mask.
 *    \    Causes the character that follows to be displayed as a literal character.
 *    Used to display any of the characters listed in this table as literal characters
 *    (for example, \A is displayed as just A).
 *    \nnnn  Single character literal (e.g. \002, \0x0A, \u2030)
 *    ''    Encloses a literal expression (for example, the pattern
 *    "990' units sold'" would display as "27 units sold")
 *    **    Encloses a password encrypted string (for example, the pattern
 *    "*AAAAaaaa* would accept a password at least 4, and at most 8
 *    alphanumeric characters.  They would echo as '*' as they were
 *    typed.
 *    {}    Can be used to bracket optional fields
 *    (e.g. "LLLL{LLL}" will allow a user to edit
 *    at least (but optionally up to 7) letters.
 *    ^    Sets the initial cursor position when editing
 *
 *
 * The string editmask can actually consist of up to 4 distinct subfields, separated by the semicolons.
 * They are:
 *     (1) The edit mask itself
 *     (2) A '0' or '1' indicating whether literals should be stripped ('0' = strip out literals)
 *     (3) The character to use as a "blank" indicator.  This is used to show the user which elements
 *   have not yet been entered.
 *     (4) Is the character to be used to replace blank characters on output. If there is no character
 *   given, they will be stripped.
 *
 * For example, the edit mask "(999)000-0000;0;_;" would indicate:
 *     (1) The template the user would see on field entry would be "(___)___-____
 *     (2) Unfilled characters will be shown by "_" (underscore)
 *     (3) Literals will be removed on output, so (408)555-1234 would become 4085551234
 *     (4) Blank characters will be stripped, so (___)555-1234 would become 5551234
 *
 * Not all 3 subfields are required.
 * If subfield 3 is omitted, the '_' (underscore) character will be the blank indicator
 * If subfield 2 is omitted, then literals will NOT be removed
 *
 */
public class ItemFormatStr extends  ItemFormatter {
  VariantFormatStr vFormatter;

  /**
   * This is the constructor to build a string-based implementation
   * of the Formatter interface.
   *
   * @param pattern Is the string of special characters used to format
   * values when using the format() method of this interface.  If this
   * value is null (or empty), the best "default" pattern will be selected based
   * on the locale.
   *
   * @param formatterType Is the type of pattern being used.  This can be any
   * one of the following 4 types: ItemFormatter.NUMERIC, ItemFormatter.DATETIME,
   * ItemFormatter.TEXT, or ItemFormatter.DECIMAL.  A value of zero will default
   * to ItemFormatter.TEXT.
   *
   * @param locale Is the locale to control this pattern.  This will determine
   * things like the decimal point sign, the currency sign, etc.  If this value
   * is null, the current default locale will be used.
   */
  public ItemFormatStr(String pattern, int variantType, Locale locale) {
    vFormatter = new VariantFormatStr(pattern, variantType, locale);
  }

  public ItemFormatStr(String pattern, int variantType) {
    this(pattern, variantType, null);
  }

  // ------------ Extension Methods for ItemFormatter -------------------
/**
* Constructs a String representing the given
* value stored in the supplied Variant.  All reasonable attempts will
* be made to "cast" the type found in the variant into the appropriate
* type specified in the constructor of the implementors of this.
*
* @param value Contains the value to be formatted.  It will be cast to
* the appropriate type where possible
*
* @return A String containing the formatted data.  An empty
* string indicates a null or empty input value.  A null return means
* the formatting failed.  Be ready for that null possibility.
*/
  public String format(Object value) throws InvalidFormatException {
    if (!(value instanceof Variant))
      throw new InvalidFormatException(Res.getString(Res.MustBeVariant));
    return vFormatter.format((Variant) value);
  }

/**
* Parses (i.e. analyzes) the given String and produces as
* output a Object containing the appropriate value.
*
* @param stringValue Contains the string to be parsed.  A null or empty
* stringValue will return a Variant.ASSIGNED_NULL variant.
*
*/
  public Object parse(String stringValue) throws InvalidFormatException {
    Variant v = new Variant();
    vFormatter.parse(stringValue, v);
    return (Object) v;
  }

  /**
   * This method will return the pattern currently being used by this
   * ItemFormatStr object for formatting and parsing.
   */
  public String getPattern() {
    return vFormatter.getPattern();
  }

  /**
   * This method will set the current pattern used by this ItemFormatStr
   * object for formatting and parsing.
   *
   * @param pattern Contains the new patter to use.  If it is null or
   * empty, this routine will automatically select the best default it
   * can based on the current locale and format type.
   *
   * @return The prior pattern (also available via getPattern()) is returned.
   * This is a convenience in case a temporary pattern is being set (to be
   * reset a short time later)
   */
  public String setPattern(String pattern) {
    return vFormatter.setPattern(pattern);
  }

  /**
   * Sets the "special" item associated with a particular ItemFormatter.
   * This is a general purpose routine to set particular booleans, characters,
   * flags etc inside a formatter -- but it is completely dependent on the
   * formatter being used.
   *
   * @param objType contains an identifier telling which object in which formatter to set.  Currently may be one of:
   *
   *   ItemFormatter.FILLCHARACTER  [Character] For Text formatters, the char to place in empty slots in the string
   *   ItemFormatter.REPLACECHARACTER  [Character] For Text formatters, the char used to replace FILLCHARACTER on output
   *
   *
   * @param obj Contains the object to be set.  The type of the Object MUST match the expected type for the
   * given 'objType'.  Do NOT pass in a null object.
   *
   * @return The object containing the prior value of this item.  Useful for restoring to original value after
   * temporary alteration.
   */
  public Object setSpecialObject(int objType, Object obj) {
    return vFormatter.setSpecialObject(objType, obj);
  }

  /**
   * Retrieves the "special" item associated with a particular ItemFormatter.
   * This is a general purpose routine to get particular booleans, characters,
   * flags etc inside a formatter -- but it is completely dependent on the
   * formatter being used.
   *
   * @param objType contains an identifier telling which object in which formatter to retrieve.  Currently may be one of:
   *
   * ItemFormatter.FILLCHARACTER  [Character] For Text formatters, the char to place in empty slots in the string
   * ItemFormatter.REPLACECHARACTER  [Character] For Text formatters, the char used to replace FILLCHARACTER on output
   *
   *
   * @return The object containing the value of this item.  temporary alteration.
   */
  public Object getSpecialObject(int objType) {
    return vFormatter.getSpecialObject(objType);
  }

  /**
   * Returns the Locale being used by this ItemFormatter.
   *
   * @return The Locale used by the VariantFormatter (will never be null)
   */
  public Locale getLocale() {
    return vFormatter.getLocale();
  }

  /**
   * Returns the JDK Format subclass associated with this ItemFormatter.
   *
   * @return The JDK Format subclass for this ItemFormatter.  Could return null.
   */
  public Format getFormatObj(){
    return vFormatter.getFormatObj();
  }
}
