/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */

package borland.jbcl.model;

import java.util.*;
import java.text.*;
import java.math.*;
import java.sql.*;

import borland.jbcl.util.*;

/**
 * This class extends the VariantFormatter class through the use of
 * String patterns to control formatting and parsing.  Though other
 * implementations of Formatter are allowed, VariantFormatStr is the basic
 * one (and only one) built into the product at this time.
 *
 * There are 4 different kinds of pattern strings which can be used.
 * Each is distinct, and the fields from one cannot be used with another.
 * The type used will be inferred from the Variant.type passed into the
 * constructor.
 * The three types are:
 *     ItemFormatter.NUMERIC  (double)
 *     ItemFormatter.DECIMAL  (BigDecimal)
 *     ItemFormatter.DATETIME
 *     ItemFormatter.TEXT
 *
 *
 * Following are detailed discussions of each of the characters allowed in
 * each of these patterns.
 *
 * For Numeric fields
 *
 *  Symbol  Meaning          Notes
 *    0    a digit
 *    #    a digit, zero shows as absent
 *    .    placeholder for decimal separator
 *    ,    placeholder for grouping delimiter.  Shows the interval to be used.
 *    ;    separates formats.                       There are two: positive and negative.
 *    %    divide by 100 and show as percentage
 *    X    any other characters can be used in
 *    -    a leading minus sign will appear for
 *    negative numbers
 *    ()    Alternatively, the entire expression
 *    can be enclosed in parentheses to show
 *    as negative (e.g. "($#,###.##)")
 *    {}    Can be used to bracket optional fields
 *    (e.g. "0000.{00}" will allow the decimal
 *    fraction to be omitted when editing
 *    ^    Sets the initial cursor position when editing
 *    \nnnn  Single character literal (e.g. \002, \0x0A, \u2030)
 *
 *  The numeric format mask actually consist of 2 semicolon separated masks.
 *  The first is required.  The second, if provided, will determine how negative
 *  numbers are formatted.  For example, "###.##;(###.##)" will format negative
 *  numbers in parentheses.
 *
 *
 *  For Date or TimeStamp fields:
 *
 *   Symbol  Meaning      Presentation  Notes
 *
 *    G    era designator    (Text)    AD/BC
 *    y    year       (Number)   Y,YY = 97, YYYY = 1997
 *    M    month in year    (Text & Number) M,MM = numeric, MMM = month abbrev, MMMM = full month
 *    d    day in month    (Number)  d,dd = 10, dddd = 0010
 *    h    hour in am/pm (1~12)  (Number)  h,hh = 12, hhhh = 0012
 *    H    hour in day (0~23)  (Number)  H,HH = 23, HHHH = 0023
 *    m    minute in hour    (Number)  m,mm = 59, mmmm = 0059
 *    s    second in     (Number)  s,ss = 59, ssss = 0059
 *    S    millisecond    (Number)  S = 9, SS = 99, SSS = 999
 *    E    day in week    (Text)    E,EE,EEE = Sun, EEEE = Sunday
 *    D    day in year    (Number)  D,DD = 364, DDDD = 0364
 *    F    day of week in month  (Number)   2 (2nd Wed in July)
 *    w    week in year    (Number)  27
 *    W    week in month    (Number)   2
 *    a    am/pm marker    (Text)     AM/PM
 *    k    hour in day (1~24)  (Number)  24
 *    K    hour in am/pm (0~11)  (Number)  0
 *    z    time zone    (Text)     z,zz,zzz = PDT, zzzz = Pacific Standard Time
 *    '    escape for text
 *    ''   single quote
 *    \nnnn  Single character literal (e.g. \002, \0x0A, \u2030)
 *    {}    Can be used to bracket optional fields
 *    (e.g. "0000.{00}" will allow the decimal
 *    fraction to be omitted when editing
 *    ^    Sets the initial cursor position when editing
 *
 * For string fields:
 * Symbol  Meaning
 *
 *    0    Digit (0 through 9, entry required; plus [+] and minus [-] signs not allowed).
 *    9    Digit or space (entry not required; plus and minus signs not allowed).
 *    #    Digit or space (entry not required; blank positions converted to spaces, plus and minus signs allowed).
 *    L    Letter (A through Z, entry required).
 *    l    Letter (A through Z, entry optional).
 *    ?    Letter (A through Z, entry optional).
 *    A    Letter or digit (entry required).
 *    a    Letter or digit (entry optional).
 *    C    Any character or a space (entry optional).
 *    c    Any character or a space (entry required).
 *    &    Any character or a space (entry required).
 *    <    Causes all characters that follow to be converted to lowercase.
 *    >    Causes all characters that follow to be converted to uppercase.
 *    !    Causes input mask to fill from right to left, rather than from left to right,
 *    when characters on the left side of the input mask are optional. You can include
 *    the exclamation point anywhere in the input mask.
 *    \    Causes the character that follows to be displayed as a literal character.
 *    Used to display any of the characters listed in this table as literal characters
 *    (for example, \A is displayed as just A).
 *    \nnnn  Single character literal (e.g. \002, \0x0A, \u2030)
 *    ''    Encloses a literal expression (for example, the pattern
 *    "990' units sold'" would display as "27 units sold")
 *    **    Encloses a password encrypted string (for example, the pattern
 *    "*AAAAaaaa* would accept a password at least 4, and at most 8
 *    alphanumeric characters.  They would echo as '*' as they were
 *    typed.
 *    {}    Can be used to bracket optional fields
 *    (e.g. "LLLL{LLL}" will allow a user to edit
 *    at least (but optionally up to 7) letters.
 *    ^    Sets the initial cursor position when editing
 *
 *
 * The string editmask can actually consist of up to 4 distinct subfields, separated by the semicolons.
 * They are:
 *     (1) The edit mask itself
 *     (2) A '0' or '1' indicating whether literals should be stripped ('0' = strip out literals)
 *     (3) The character to use as a "blank" indicator.  This is used to show the user which elements
 *   have not yet been entered.
 *     (4) Is the character to be used to replace blank characters on output. If there is no character
 *   given, they will be stripped.
 *
 * For example, the edit mask "(999)000-0000;0;_;" would indicate:
 *     (1) The template the user would see on field entry would be "(___)___-____
 *     (2) Unfilled characters will be shown by "_" (underscore)
 *     (3) Literals will be removed on output, so (408)555-1234 would become 4085551234
 *     (4) Blank characters will be stripped, so (___)555-1234 would become 5551234
 *
 * Not all 3 subfields are required.
 * If subfield 3 is omitted, the '_' (underscore) character will be the blank indicator
 * If subfield 2 is omitted, then literals will NOT be removed
 *
 */
public class VariantFormatStr extends VariantFormatter {
  String pattern;
  ParsePosition parsePos;
  FieldPosition fieldPos;
  DecimalFormat decimalFormat;
  SimpleDateFormat timeFormat;
  TextFormat textFormat;
  BooleanFormat booleanFormat;
  Locale locale;
  int variantType;
  int formatterType;
  int scale;
  int precision;
  boolean isCurrency;

  /**
   * This is the constructor to build a string-based implementation
   * of the Formatter interface.
   *
   * @param pattern Is the string of special characters used to format
   * values when using the format() method of this interface.  If this
   * value is null (or empty), the best "default" pattern will be selected based
   * on the locale.
   *
   * @param variantType Is the Variant data type which will be used by this
   * formatter (see the format() and parse() methods in this class).
   *
   * @param locale Is the locale to control this pattern.  This will determine
   * things like the decimal point sign, the currency sign, etc.  If this value
   * is null, the current default locale will be used.
   *
   * @param scale Is currently used only for BigDecimal data types.  Any value
   * other than -1 will select the number of decimal digits used in formatting
   * and parsing BigDecimal values.  Note that the pattern must still express
   * the number of digits to be displayed.
   *
   * @param precision Is currently unused.  Must presently be -1.
   *
   * @param isCurrency Is used to indicate that a numeric field is currency.  This
   * parameter is used only if the pattern is left blank (which asks this constructor
   * to manufacture a pattern from the given Locale).  A true value will select the
   * Locale's currency format.  A false will select a non-currency numeric format.
   */
public VariantFormatStr(String pattern,
                        int variantType,
                        Locale locale,
                        int scale,
                        int precision,
                        boolean isCurrency) {

//Diagnostic.addTraceCategory(Trace.FormatStr);

    this.scale = scale;
    this.precision = precision;
    this.isCurrency = isCurrency;


    //
    // We tolerate a bad variant type and revert to "String" type.  This
    // allows controls which do not know what type they want to at least
    // get a reasonable default.  Data-aware controls know their type.
    //
    if (variantType <= Variant.NULL_TYPES /* || variantType > Variant.MaxTypes*/)
      variantType = Variant.STRING;    

    this.variantType = variantType;
    this.formatterType = formatTypeFromVariantType(variantType);

    if (locale == null)
      this.locale = Locale.getDefault();
    else this.locale = locale;

    //
    // The following objects are required for the JDK Format subclasses
    // (we are essentially a layer on top of them)
    //
    parsePos = new ParsePosition(0);
//    fieldPos = new FieldPosition(0);
    decimalFormat = null;
    timeFormat = null;
    textFormat = null;
    booleanFormat = null;
    //
    // A null pattern means we load the best default we can from those
    // in LocaleElements based on locale.  This is actually a popular path
    // through the code, since it allows columns to track their locales
    // as apps move to different countries.
    //
    this.pattern = (pattern == null || pattern.length() == 0)
                   ? getDefaultPattern(this.variantType)
                   : (formatterType == VariantFormatter.TEXT)
                      ? pattern
                      : buildTrueFormatMask(pattern);
    Diagnostic.trace(Trace.FormatStr, "VariantFormatStr using pattern \"" + this.pattern + "\"");

    //
    // Now that we know we have a pattern, create the JDK Format object
    // which can deal with this type of pattern.
    //
    Diagnostic.trace(Trace.FormatStr, "VariantFormatStr constructor using formatterType = " + formatterType + " (variant " + variantType + ")");
    switch (formatterType)
    {
      case ItemFormatter.DECIMAL:
  // falls into
      case ItemFormatter.NUMERIC:
        decimalFormat = (DecimalFormat) NumberFormat.getNumberInstance(this.locale);
//        try {
// ktien: We don't currently support localized patterns.
//          decimalFormat.applyLocalizedPattern(this.pattern);
          decimalFormat.applyPattern(this.pattern);
//        }
//        catch (ParseException pe) {
//        }
        break;

      case ItemFormatter.DATETIME:
  //timeFormat = (SimpleDateFormat) DateFormat.getDateTimeFormat(DateFormat.DEFAULT, DateFormat.DEFAULT, this.locale);
        timeFormat = (SimpleDateFormat) DateFormat.getDateTimeInstance(DateFormat.DEFAULT, DateFormat.DEFAULT, this.locale);

        // ktien: 
        // Above factory method sets the timezone to the default for the given locale.
        // A given locale can have multiple timezones (e.g. in the US),
        // so the returned timezone will not necessary match the user's time zone.
        // There is no DateFormat constructor/factory method that takes a timezone -- need to set
        // after construction.  Following will set the timezone to the default.  We probably
        // should, in the next version, add a constructor for VariantFormatStr that takes
        // a timezone.
        timeFormat.setTimeZone(TimeZone.getDefault());

        Diagnostic.trace(Trace.FormatStr, "At construction, timeFormat.setPattern(" + pattern + ")");

//ktien
//        timeFormat.applyLocalizedPattern(this.pattern);
        timeFormat.applyPattern(this.pattern);
        break;

      case ItemFormatter.BOOLEAN:
        booleanFormat = new BooleanFormat(this.pattern);
        break;

      case ItemFormatter.TEXT:
        default:
          Diagnostic.trace(Trace.FormatStr, "VariantFormatStr defaulting to text");
          textFormat = new TextFormat(this.pattern);
          break;
    }
  }  // end of VariantFormatStr constructor

  /**
   * This is the constructor to build a string-based implementation
   * of the Formatter interface.
   *
   * @param pattern Is the string of special characters used to format
   * values when using the format() method of this interface.  If this
   * value is null (or empty), the best "default" pattern will be selected based
   * on the locale.
   *
   * @param variantType Is the Variant data type which will be used by this
   * formatter (see the format() and parse() methods in this class).
   *
   */
  public VariantFormatStr(String pattern, int variantType) {
    this(pattern, variantType, null);
  }

  /**
   * This is the constructor to build a string-based implementation
   * of the Formatter interface.
   *
   * @param pattern Is the string of special characters used to format
   * values when using the format() method of this interface.  If this
   * value is null (or empty), the best "default" pattern will be selected based
   * on the locale.
   *
   * @param variantType Is the Variant data type which will be used by this
   * formatter (see the format() and parse() methods in this class).
   *
   * @param locale Is the locale to control this pattern.  This will determine
   * things like the decimal point sign, the currency sign, etc.  If this value
   * is null, the current default locale will be used.
   */
  public VariantFormatStr(String pattern, int variantType, Locale locale) {
    this(pattern, variantType, locale, -1, -1, false);
}

// --------------- Extension Methods for VariantFormatter ----------------------

  /**
   *  Returns the Variant type this VariantFormatter will use for its getValue().
   *
   * @return The Variant type used to control the Variant type in getValue();
   */
  public int getVariantType() {
    return variantType;
  }

  /**
   *  This method will format (i.e. produce a String) from the given
   *  value using the current pattern.
   *
   *  @param value Contains the value to be formatted. The type of data
   *  stored in the Variant is self-identifying via value.getType().  Not
   *  all types are supported, but every reasonable attempt will be made
   *  to cast the type provided into the type required for the format logic.
   *  Note that the value Variant will not be altered by this function.
   *
   *  @return A String containing the formatted data.  This String
   *  could be empty if the input Variant was null or unassigned.
   *  A null will be returned in the event formatting fails.
   */
  public String format(Variant value) {
    StringBuffer result = null;

    //
    // It is acceptable to pass in nulls or empty variants -- but we guarantee an empty string out
    // The only exception is boolean formatting which may want a special value for an null

    if (booleanFormat == null && (value == null || value.isNull() || !(value instanceof Variant)))
      return new String();

    fieldPos = new FieldPosition(0);
//    fieldPos.alignField = 0;  // The JDK Format objects use this information -- init it each time
//    fieldPos.setBeginIndex(0);
//    fieldPos.setEndIndex(0);

    //
    // Only one of the Format objects will be defined based on the type we defined in the constructor.
    // Simply decode the right one.
    //
    if (timeFormat != null) {
      java.util.Date dateObj = variantToLongDate(value);    // SimpleDateFormat always wants Dates
      if (dateObj != null) {
        try {
          result = new StringBuffer();
          Diagnostic.trace(Trace.FormatStr, "timeFormat.format(" + dateObj.toString() + ")");
          result = timeFormat.format(dateObj, result, fieldPos);
        }
        catch (IllegalArgumentException e) {
          result = null;
        }
      }
    }

    else if (decimalFormat != null) {
      Double doubleObj = variantToDouble(value, scale);  // DecimalFormat always wants doubles
      if (doubleObj != null) {
        result = new StringBuffer();
        result = decimalFormat.format(doubleObj.doubleValue(), result, fieldPos);
      }
    }

    else if (textFormat != null) {
      String stringObj = variantToString(value);
      if (stringObj != null) {
        result = new StringBuffer();
        result = textFormat.format(stringObj, result, fieldPos);
      }
    }
    else if (booleanFormat != null) {
      Boolean booleanObj = variantToBoolean(value);
      result = new StringBuffer();
      result = booleanFormat.format(booleanObj, result, fieldPos);
    }

    return result == null ? null : result.toString();
  }

  /**
   * This method will parse a string using the current pattern and produce
   * the appropriate value which is returned in the form of a Variant.
   *
   * @param stringValue Contains the string to be parsed.  A null or empty stringValue
   * will result in a Variant object being returned which is set to an AssignedNull
   *
   * @param value Is a pre-allocated Variant used to contain the result.
   * Its type will be determined by the 'variantType' parameter.
   *
   * @param variantType Contains one of the Variant.type values to control
   * the output type of the returned Variant
   *
   */
  private static boolean BUG6296 = true;
  private static boolean BUG5595 = true;
                               
  public void parse(String stringValue, Variant value, int variantType) throws InvalidFormatException {

    if (value == null || !(value instanceof Variant)) {
      Diagnostic.println(value + " is not a variant in VariantFormatStr.parse()");
      if (value != null)
        Diagnostic.println("  it is of class: " + value.getClass().getName());
      throw new InvalidFormatException(Res.getString(Res.MustBeVariant));
    }
    //
    // If no string was given (or it is empty), turn it right back
    // around as an AssignedNull.  Don't rely on the parse routine to
    // handle this.
    //
    if (stringValue == null || stringValue.length() == 0) {
      value.setNull(Variant.ASSIGNED_NULL);
      return;
    }

    Diagnostic.trace(Trace.FormatStr, "FormatStr.parse(\"" + stringValue + "\", " + Variant.typeName(variantType));

    //
    // Make sure the variantType passed in is valid -- use the default if questionable
    //
    if (variantType <= Variant.NULL_TYPES)
      variantType = this.variantType;

    parsePos.setIndex(0);    // JDK parse routines use this to know where to start and also
                             // return the parse failure point here

    if (timeFormat != null) {
      Diagnostic.trace(Trace.FormatStr, "Parsing date " + stringValue + " using pattern " + timeFormat.toLocalizedPattern());

      java.util.Date dateResult = timeFormat.parse(stringValue, parsePos);
      if (dateResult == null) {
        Diagnostic.println(" timeFormat returned null");
        throw new InvalidFormatException(Res.getString(Res.ParseError), parsePos.getIndex());
      }
      Diagnostic.trace(Trace.FormatStr, "which yielded " + dateResult + " of class " + dateResult.getClass().getName());
      Diagnostic.trace(Trace.FormatStr, "parse() calling longDateToVariant");

      value = longDateToVariant(dateResult, value, variantType);
    }

    else if (decimalFormat != null) {
      Number nResult = decimalFormat.parse(stringValue, parsePos);

      if (nResult == null) {
        try {
          if (BUG6296 && stringValue.charAt(0) == '+') {
//System.err.println("invoking bug 6296 workaround...");
            FastStringBuffer fs = new FastStringBuffer(stringValue);
            fs.removeCharAt(0);
            try {
              nResult = decimalFormat.parse(fs.toString(), parsePos);
            }
            catch (Exception ex) {
//System.err.println(" still cannot parse");              
            }
          }
      // when parsing a zero with ###.  We get around this by trying a simpler parse
      // technique (but only when we fail for this specific case)
          else if (BUG5595) {
            Double d = Double.valueOf(stringValue);   // will throw if has a problem parsing
            if (d.doubleValue() == 0.0) {             // limit this to zero value case
              nResult = d;
              parsePos.setIndex(1);
            }
          }
        }
//        catch (InvalidFormatException ex) {
//System.err.println(" rethrowing invalidFormatException");          
//          throw ex;
//        }
        catch (Exception ex) {
//ex.printStackTrace();          
        }
      }

        if (nResult == null ||
          (nResult instanceof Double && ((Double)nResult).isNaN()) ||
          parsePos.getIndex() == 0)
        value = null;
      if (nResult != null) Diagnostic.trace(Trace.FormatStr, "decimal.parse gives: " + nResult
                                              + " and doubleVal as " + nResult.doubleValue());  //NORES
      if (nResult != null) doubleToVariant(nResult.doubleValue(), value, variantType, scale);
    }

    else if (textFormat != null) {
      StringBuffer sResult = textFormat.parse(stringValue, parsePos);
      if (sResult == null)
        value = null;
      else value = stringToVariant(sResult.toString(), value, variantType, scale);
    }

    else if (booleanFormat != null) {
      Boolean booleanObj = booleanFormat.parse(stringValue, parsePos);
      // To distinguish between a null meaning a legal null string vs something
      // we cannot parse, we check whether the parsePosition was advanced.
      // An empty string is legal for the nullString value.
      if (booleanObj == null && stringValue != null && stringValue.length() > 0 && parsePos.getIndex() == 0)
        throw new InvalidFormatException(Res.getString(Res.ParseError), parsePos.getIndex());
      value = booleanToVariant(booleanObj, value, variantType);
    }

    else value = null;    // should never get here, but this is for robustness

    if (value == null)
      throw new InvalidFormatException(Res.getString(Res.ParseError), parsePos.getIndex());
  }

  /**
   * This method will parse a string using the current pattern and produce
   * the appropriate value which is returned in the form of a Variant.
   *
   * @param stringValue Contains the string to be parsed.  A null or empty stringValue
   * will result in a Variant object being returned which is set to an AssignedNull
   *
   * @param value Is a pre-allocated Variant used to contain the result.
   * Its type will be determined by the 'variantType' parameter originally
   * passed into the constructor of this ItemFormatter.
   *
   */
  public void parse(String stringValue, Variant value) throws InvalidFormatException {
    parse(stringValue, value, this.variantType);
  }

  /**
   * This method will return the pattern currently being used by this
   * VariantFormatStr object for formatting and parsing.
   */
  public String getPattern() {


/*
 *    String pattern = (decimalFormat != null)
 *           ? decimalFormat.getPattern(true)
 *           : (timeFormat != null)
 *       ? timeFormat.getPattern(true)
 *       : textFormat.getPattern();
 */
    Diagnostic.trace(Trace.FormatStr, "VariantFormatStr: getPattern returning \"" + pattern);
    return pattern;
  }

  /**
   * This method will set the current pattern used by this VariantFormatStr
   * object for formatting and parsing.
   *
   * @param pattern Contains the new patter to use.  If it is null or
   * empty, this routine will automatically select the best default it
   * can based on the current locale and format type.
   *
   * @return The prior pattern (also available via getPattern()) is returned.
   * This is a convenience in case a temporary pattern is being set (to be
   * reset a short time later)
   */
  public String setPattern(String pattern) {
    String oldFormat = getPattern();
    if (pattern == null || pattern.length() == 0)
      pattern = getDefaultPattern(variantType);

    Diagnostic.trace(Trace.FormatStr, "VariantFormatStr: setting pattern to \"" + pattern + "\"");

//    try {
      if (decimalFormat != null)
//ktien
//        decimalFormat.applyLocalizedPattern(pattern);
        decimalFormat.applyPattern(pattern);
      else if (timeFormat != null)
//        timeFormat.applyLocalizedPattern(pattern);
        timeFormat.applyPattern(pattern);
      else if (textFormat != null)
        textFormat.applyPattern(pattern);
      else if (booleanFormat != null)
        booleanFormat.applyPattern(pattern);
      this.pattern = pattern;
//    }
//    catch (ParseException pe) {
//      
//    }
    return oldFormat;
  }

  /**
   * Sets the "special" item associated with a particular ItemFormatter.
   * This is a general purpose routine to set particular booleans, characters,
   * flags etc inside a formatter -- but it is completely dependent on the
   * formatter being used.
   *
   * @param objType contains an identifier telling which object in which formatter to set.  Currently may be one of:
   *
   *   ItemFormatter.FILLCHARACTER  [Character] For Text formatters, the char to place in empty slots in the string
   *   ItemFormatter.REPLACECHARACTER  [Character] For Text formatters, the char used to replace FILLCHARACTER on output
   *
   *
   * @param obj Contains the object to be set.  The type of the Object MUST match the expected type for the
   * given 'objType'.  Do NOT pass in a null object.
   *
   * @return The object containing the prior value of this item.  Useful for restoring to original value after
   * temporary alteration.
   */
  public Object setSpecialObject(int objType, Object obj) {
    Object resultObj = null;

    if (textFormat != null) {
      switch (objType)
      {
        case ItemFormatter.FILLCHARACTER:
          resultObj = new Character(textFormat.getFillCharacter());
          textFormat.setFillCharacter(((Character)obj).charValue());
          break;
        case ItemFormatter.REPLACECHARACTER:
          resultObj = new Character(textFormat.getReplaceCharacter());
          textFormat.setReplaceCharacter(((Character)obj).charValue());
          break;
      }
    }
    return resultObj;
  }

  /**
   * Retrieves the "special" item associated with a particular ItemFormatter.
   * This is a general purpose routine to get particular booleans, characters,
   * flags etc inside a formatter -- but it is completely dependent on the
   * formatter being used.
   *
   * @param objType contains an identifier telling which object in which formatter to retrieve.  Currently may be one of:
   *
   * ItemFormatter.FILLCHARACTER  [Character] For Text formatters, the char to place in empty slots in the string
   * ItemFormatter.REPLACECHARACTER  [Character] For Text formatters, the char used to replace FILLCHARACTER on output
   *
   *
   * @return The object containing the value of this item.  temporary alteration.
   */
  public Object getSpecialObject(int objType) {

    if (textFormat != null) {
      switch (objType)
      {
        case ItemFormatter.FILLCHARACTER:
          return new Character(textFormat.getFillCharacter());
        case ItemFormatter.REPLACECHARACTER:
          return new Character(textFormat.getReplaceCharacter());
      }
    }
    return null;
  }

  /**
   * Returns the Locale being used by this ItemFormatter.
   *
   * @return The Locale used by the VariantFormatter (will never be null)
   */
  public Locale getLocale() {
    return locale;
  }

  /**
   * Returns the JDK Format subclass associated with this ItemFormatter.
   *
   * @return The JDK Format subclass for this ItemFormatter.  Could return null.
   */
  public Format getFormatObj(){
    return (decimalFormat != null)
      ? (Format) decimalFormat
      : (timeFormat != null)
        ? (Format) timeFormat
        : (textFormat != null)
           ? (Format) textFormat
           : (booleanFormat != null)
              ? (Format) booleanFormat
              : null;
  }

  /**
   * Returns the current scale factor
   */
  public int getScale() { return scale; };

  /**
   *  Strips out any Borland-specific extensions from an edit/format mask.
   *
   *  @param editMask Contains the editMask to be used for formatting/parsing/editing
   *
   *  @return A new String stripped of any Borland extensions to the JDK's Format strings
   *
   */
  public static final String buildTrueFormatMask(String editMask) {
    int iLen = editMask.length();
    char[] src = new char[iLen];
    char[] dst = new char[iLen];
    char c;

    // Currently, merely stripping the '{','}' and '^' are sufficient to
    // make a Borland edit mask into a legal JDK edit mask.
    //
    editMask.getChars(0, iLen, src, 0);

    int iDst = 0;
    for (int i = 0; i < iLen; ++i) {
      c = src[i];
      switch (c)
      {
        case '^':
        case '{':
        case '}':
          Diagnostic.trace(Trace.FormatStr, "BuildTrueFormat: stripping \'" + c + "\'");
          continue;
        default:
          dst[iDst++] = src[i];
      }
    }

    return new String(dst, 0, iDst);
  }


// ================== internal methods ======================================

  //
  // This method will return one of the basic 4 ItemFormatter.types based on given variant type
  //
  static final int formatTypeFromVariantType(int variantType) {
    switch (variantType)
    {
      case Variant.BYTE:
      case Variant.SHORT:
      case Variant.INT:
      case Variant.LONG:
      case Variant.FLOAT:
      case Variant.DOUBLE:
        return ItemFormatter.NUMERIC;

      case Variant.BIGDECIMAL:
        return ItemFormatter.DECIMAL;

      case Variant.BOOLEAN:
        return ItemFormatter.BOOLEAN;

      case Variant.DATE:
      case Variant.TIME:
      case Variant.TIMESTAMP:
        return ItemFormatter.DATETIME;
      default: break; // to make compiler happy
    }
    return ItemFormatter.TEXT;
  }

  //
  // This method will determine the best "default" pattern to use based
  // on the current format type using the current Locale.  It is a good
  // idea to pass null patterns to the formatter so that you get the
  // default locale formatting when apps move to different locales.
  //
  protected String getDefaultPattern(int variantType) {
    String pattern = null;
    String[] resourceArray;
    int offset = 0;
    Diagnostic.trace(Trace.FormatStr, "VariantFormatStr.getDefaultPattern(" + Variant.typeName(variantType) + "' " + locale.toString() + ")");
    try {
//      ResourceBundle resource = ResourceBundle.load(new TextLoader(), "LocaleElements", locale);
//      ResourceBundle resource =
//        ResourceBundle.getBundle("java.text.resources.LocaleElements", locale,
//                                         getClass().getClassLoader());//DateFormat.class.getClassLoader());
      ResourceBundle resource = SystemResourceBundle.getBundle("java.text.resources.LocaleElements", locale);

      switch (formatterType)
      {
        case ItemFormatter.DECIMAL:
          //System.err.println("DECIMAL default pattern");
          // falls into NUMERIC
        case ItemFormatter.NUMERIC:

          switch (variantType) {
            case Variant.BYTE:
              pattern = new String("###");
              break;
            case Variant.SHORT:
              pattern = new String("#####");
              break;
            case Variant.INT:
              pattern = new String("###########");
              break;
            case Variant.LONG:
              pattern = new String("####################");
              break;
            default:
              //case Variant.BIGDECIMAL:
              //case Variant.DOUBLE:
              resourceArray = resource.getStringArray("NumberPatterns");
              if (isCurrency) {             // if user asked for currency, use currency pattern
                //System.err.println(" using currency");
                ++offset;
              }
              pattern = resourceArray[offset];
              //System.err.println(" pattern is: " + pattern);
              break;
          }
          break;

        case ItemFormatter.DATETIME:
          resourceArray = resource.getStringArray("DateTimePatterns");

          switch (variantType) {
            case Variant.TIME:
              pattern = resourceArray[2];    // medium time
              break;
            case Variant.TIMESTAMP:
              //System.err.println("resourceArray size is: " + resourceArray.length);
              //System.err.println("Default timestamp, msgFormat pattern is: " + resourceArray[8]);
              //System.err.println("  params are: " + resourceArray[7] + " and " + resourceArray[2]);

              // JDK now provides a resourced string to dictate how date + time are combined
              pattern = MessageFormat.format(resourceArray[8], new Object[] {resourceArray[2], resourceArray[7]});
//              pattern = new String(resourceArray[7] + " " + resourceArray[2]);
              //System.err.println("  yields timeStamp format of: " + pattern);
              break;        // medium date + medium time
            default:
              //case Variant.Date:
              pattern = resourceArray[7];  // short date
              break;
          }
          break;

        case ItemFormatter.BOOLEAN:
          pattern = Res.getString(Res.TrueFalsePattern);
          break;

        case ItemFormatter.TEXT:
        default:
          pattern = new String("");
          // String with no pattern is an anomaly triggered in the general case by our painters asking
          // for a formatter for all types.  The empty string has special signficance in string formatting/parsing
          break;
      }  // end switch
    }    // end try
    catch (MissingResourceException e) {  // should NEVER get here, but if we do, use US symbols
      pattern = new String("");
    }

    Diagnostic.trace(Trace.FormatStr, "VariantFormatStr: (" + formatterType + "/" + Variant.typeName(variantType) +
                       ") chose default pattern of \"" + pattern + "\"");
    return pattern;
  }

  //
  // This method will produce a Date object out of the given variant if at
  // all possible.  It will return a null for uncastable variants.
  //
  static final java.util.Date variantToLongDate(Variant value) {

    java.util.Date result = null;

    switch (value.getType())
    {
      case Variant.LONG:
        result = new java.util.Date(value.getLong());
        break;

      case Variant.DATE:
        java.sql.Date dateObj = value.getDate();
        result = new java.util.Date(dateObj.getTime());
        break;

      case Variant.TIME:
        Time timeObj = value.getTime();
        result = new java.util.Date(timeObj.getTime());
        break;

      case Variant.TIMESTAMP:
        Timestamp timestampObj = value.getTimestamp();
        //result = new java.util.Date(timestampObj.getTime());
        result = new java.util.Date(timestampObj.getTime() + timestampObj.getNanos() / 1000000);
          // Note: as of 2/1/97, JDK stores fractional seconds in 'nanos' and forces the client
          // to add them in as needed.

        break;

      default:
        result = new java.util.Date(value.toString());  // fallback -- not expected to use this path
    }
    Diagnostic.trace(Trace.FormatStr, "variantToLongDate --> " + result);
    return result;
  }

  //
  // This method is the complement of the one above.  It will take
  // a long value and place it into the given variant, making all
  // reasonable attempts to cast it into the appropriate type currently
  // in the variant. This is used strictly for date/time stuff.
  //
  static final Variant longDateToVariant(java.util.Date dateVal, Variant value, int variantType) {
    Variant newValue;

    Diagnostic.trace(Trace.FormatStr, "longDateToVariant(" + dateVal + ") -->" + dateVal.getClass().getName());

    if (value == null)    // robustness measure
      value = new Variant();

    if (dateVal == null) {  // robustness measure
      Diagnostic.trace(Trace.FormatStr, "longDateToVariant(null)");
      value.setAssignedNull();
      return value;
    }

    switch (variantType)
    {
      case Variant.DATE:
        Diagnostic.trace(Trace.FormatStr, "Date assign " + dateVal);
        if (dateVal == null) {
          Diagnostic.trace(Trace.FormatStr, "LongDateToVariant -- yikes, it's null!");
          value.setNull(Variant.ASSIGNED_NULL);
        }
        else {
          value.setDate(dateVal.getTime());
        }
        break;

      case Variant.TIME:
        Diagnostic.trace(Trace.FormatStr, "Time assign " + dateVal);
        java.sql.Time timeObj = new java.sql.Time(dateVal.getTime());
        Diagnostic.trace(Trace.FormatStr, "  assigning " + timeObj);
        value.setTime(timeObj);
        break;

      case Variant.TIMESTAMP:
         Diagnostic.trace(Trace.FormatStr, "Timestamp assign " + dateVal);
        Timestamp timestampObj = new Timestamp(dateVal.getTime());
        Diagnostic.trace(Trace.FormatStr, "  assigning " + timestampObj);
        value.setTimestamp(timestampObj);
        break;

      case Variant.LONG:
        Diagnostic.trace(Trace.FormatStr, "Long assign " + dateVal);
        value.setLong(dateVal.getTime());
        break;

      default:
        return null;  

    }
    Diagnostic.trace(Trace.EditMaskStr, "longDateToVariant returning");
    return value;
  }

  static final Boolean variantToBoolean(Variant value) {
    //System.err.println("variantToBoolean: variant is " + value);

    Boolean result = null;

    switch (value.getType())
    {
      case Variant.BOOLEAN:
        result = new Boolean(value.getBoolean());
        break;


      case Variant.BYTE:
      case Variant.SHORT:
      case Variant.INT:
        result = new Boolean((value.getAsInt() != 0));
        break;

      case Variant.LONG:
        result = new Boolean((value.getLong() != 0));
        break;
    }
    //System.err.println(" returning " + result);

    return result;
  }

  //
  // This method will make all reasonable attempts to return a Double
  // object from the given variant. It will return null if it cannot.
  //
  static final Double variantToDouble(Variant value, int scale) {

    Double result = null;

    switch (value.getType())
    {
      case Variant.FLOAT:
        result = new Double((double)value.getFloat());
        break;
      case Variant.DOUBLE:
        result = new Double(value.getDouble());
        break;

      case Variant.BIGDECIMAL:
        // Further note: this weirdness of going through a string first seems
        // to be necessary.  Converting directly to a Double loses considerable precision.
        //String s = value.getBigDecimal().toString();
        //result = new Double(s);
        BigDecimal bn = value.getBigDecimal();
        //System.err.println("variantToDouble(), scale is: " + scale);
        if (scale >= 0)
          bn = bn.setScale(scale, BigDecimal.ROUND_HALF_UP);
        result = new Double(bn.doubleValue());
        //System.err.println(" results in double: " + result);

//        result = new Double(value.getBigDecimal().doubleValue());
          // this works better in the FCS JDK
        Diagnostic.trace(Trace.FormatStr, "variantToDouble -> " + result.toString());
        break;

      case Variant.BYTE:
      case Variant.SHORT:
      case Variant.INT:
        result = new Double((double) value.getAsInt());
        break;

      case Variant.LONG:
        result = new Double((double) value.getLong());
        break;

      default:
        result = new Double(value.toString());
    }
    return result;
  }

  static final Variant booleanToVariant(Boolean bool, Variant value, int variantType) {
    if (value == null)
      value = new Variant();

    if (bool == null)
      value.setAssignedNull();

    else switch (variantType)
    {
      case Variant.BOOLEAN:
        value.setBoolean(bool.booleanValue());
        break;
    }
    return value;
  }

  // This method is a complement of the above.  It will make every
  // effort to store the given doubleValue into the given variant
  // in the current type of the variant.
  //
  static final Variant doubleToVariant(double doubleValue, Variant value, int variantType, int scale)
      throws InvalidFormatException {
    Diagnostic.trace(Trace.FormatStr, "doubleToVariant(" + doubleValue + ", " + Variant.typeName(variantType) + ")");
//System.err.println("doubleToVariant(" + doubleValue + ", " + Variant.typeName(variantType) + ")");

    if (value == null)
      value = new Variant();

    int i;
    switch (variantType)
    {
      case Variant.BYTE:
        i = (int) doubleValue;
        if (i > 127 || i < -128) {
          throw new InvalidFormatException(Res.getString(Res.BytePrecisionBad));
        }
        value.setByte((byte) doubleValue);
        break;

      case Variant.SHORT:
        i = (int) doubleValue;
        if (i > 32767 || i < -32768) {
          throw new InvalidFormatException(Res.getString(Res.ShortPrecisionBad));
        }
        value.setShort((short) doubleValue);
        break;

      case Variant.INT:
        value.setInt((int) doubleValue);
        break;

      case Variant.LONG:
        value.setLong((long) doubleValue);
        break;

      case Variant.DOUBLE:
        value.setDouble(doubleValue);
        break;
      case Variant.FLOAT:
        value.setFloat((float)doubleValue);
        break;

      case Variant.BIGDECIMAL:
        //String s = "" + doubleValue;
        // TODO <rac> Again, double to BigDecimal is problematic.  This business
        // of going to a string first and then to a BigDecimal seems to be the only
        // current way to avoid precision loss (1/17/97)
        //BigDecimal bn = new BigDecimal(s);
        //System.out.println("** double = " + s + " and bigDecimal = " + bn);   //$$$
        BigDecimal bn = new BigDecimal(doubleValue);
        //System.err.println("--- scale is " + scale + " ---");
        if (scale >= 0)
          bn  = bn.setScale(scale, BigDecimal.ROUND_HALF_UP);
        //System.err.println("  value is " + bn);
        value.setBigDecimal(bn);
        Diagnostic.trace(Trace.FormatStr, "DoubleToVariant -> " + bn.toString());
        break;

      default:
        return null;

    }
    return value;
  }


  //
  // This routine will return a String from the given variant
  // Be warned, it is expected that what is in the Variant is already
  // a String, and it is only common courtesy to try for a cast if
  // possible.  We cannot promise International conversion rules if
  // someone gave us a double or something.
  //
  static final String variantToString(Variant value) {

    String result;
    switch (value.getType()) {
      case Variant.STRING:
        result = value.getString();
        break;

      default:
        result = value.toString();
    }

    return result;

  }

  //
  // This method will attempt to put the given StringBuffer into the
  // given variant.  It is really intended to be used strictly by
  // ItemFormatter.TEXT type of parsing, so no other casting is attempted.
  //
  static final Variant stringToVariant(String stringVal, Variant value, int variantType, int scale) {
    switch (variantType)
    {
      case Variant.STRING:  // the most common we hope
        value.setString(stringVal.toString());
        break;

      case Variant.BYTE:
        value.setByte((byte) Integer.parseInt(stringVal));
        break;

      case Variant.SHORT:
        value.setShort((short) Integer.parseInt(stringVal));
        break;

      case Variant.INT:
        value.setInt((int) Integer.parseInt(stringVal));
        break;

      case Variant.LONG:
        value.setLong((long) Long.parseLong(stringVal));
        break;

      case Variant.FLOAT:
        Float f = new Float(stringVal);
        value.setFloat(f.floatValue());
        break;
      case Variant.DOUBLE:
        Double d = new Double(stringVal);
        value.setDouble(d.doubleValue());
        break;

      case Variant.BIGDECIMAL:
        BigDecimal bn = new BigDecimal(stringVal);
        if (scale >= 0)
          bn.setScale(scale);
        value.setBigDecimal(bn);
        Diagnostic.trace(Trace.FormatStr, "StringToVariant(" + stringVal + ") -> " + bn.toString());

        break;

      case Variant.DATE:
        java.util.Date dateObj = new java.util.Date(java.util.Date.parse(stringVal));
        value.setDate(dateObj.getTime());
        break;

      case Variant.TIME:
        java.sql.Time timeObj = new java.sql.Time(java.util.Date.parse(stringVal));
        value.setTime(timeObj);
        break;

      case Variant.TIMESTAMP:
        Timestamp timestampObj = new Timestamp(java.util.Date.parse(stringVal));
        value.setTimestamp(timestampObj);
        break;

      default:
        return null;
    }
    return value;
  }
}
