/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */
//--------------------------------------------------------------------------------------------------
// Copyright (c) 1996-1998 Borland International, Inc. All Rights Reserved.
//--------------------------------------------------------------------------------------------------

package borland.jbcl.util;

import java.util.*;

/**
 *
 */
public class EventMulticaster
{
  /**
   * High speed test to see if any listeners present.
   */
  public final boolean hasListeners() {
    return listeners != null;
  }

  /**
   * High speed dispatcher that does not need to be synchronized.
   */
  public final void dispatch(DispatchableEvent e) {
    // Synchronized not needed becuase all updates are made to a "copy"
    // of listeners.  Assumes reference assignment is atomic.
    //
    EventListener[] listenersCopy = this.listeners;

    // Once I have a local copy of the list, don't have to worry about threads
    // adding/deleting from this list since they will make a copy of the list,
    // not modify the list.
    //
    if (listenersCopy != null) {
      int count = listenersCopy.length;
      for (int index = 0; index < count; ++index) {
        // WARNING:  Please don't checkin diagnostics here.  It slows my smoke
        // test down enormously.  Test that usually takes < 2 minutes, takes 15 minutes
        // when one of the diagnostics below was left in.  (SteveS).
        //
        //long time = System.currentTimeMillis();
        //Diagnostic.trace(EventMulticaster.class, "->dispatch e=" + e + " => " + listenersCopy[index]);
        e.dispatch(listenersCopy[index]);
        //Diagnostic.trace(EventMulticaster.class, "  dispatch took " + (System.currentTimeMillis()-time) + "ms");
      }
    }
  }

  /**
   * High speed dispatcher that does not need to be synchronized.
   * Used for vetoable events (ones that throw VetoException to deny the action
   * of an event)  Returns false if an event listener throws a VetoException.
   */
  public final boolean vetoableDispatch(VetoableDispatch e) {
    // Synchronized not needed becuase all updates are made to a "copy"
    // of listeners.  Assumes reference assignment is atomic.
    EventListener[] listenersCopy = this.listeners;

    // Once I have a local copy of the list, don't have to worry about threads
    // adding/deleting from this list since they will make a copy of the list,
    // not modify the list.
    try {
      if (listenersCopy != null) {
        int count = listenersCopy.length;
        for (int index = 0; index < count; ++index) {
          e.vetoableDispatch(listenersCopy[index]);
        }
      }
    }
    catch (VetoException ex) {
      return false;
    }
    return true;
  }

  /**
   * High speed dispatcher that does not need to be synchronized.
   * Used for events that can throw an exception.
   */
  public final void exceptionDispatch(ExceptionDispatch e)
    throws Exception
  {
    // Synchronized not needed becuase all updates are made to a "copy"
    // of listeners.  Assumes reference assignment is atomic.
    EventListener[] listenersCopy = this.listeners;

    // Once I have a local copy of the list, don't have to worry about threads
    // adding/deleting from this list since they will make a copy of the list,
    // not modify the list.
      if (listenersCopy != null) {
        int count = listenersCopy.length;
        for (int index = 0; index < count; ++index) {
          e.exceptionDispatch(listenersCopy[index]);
        }
      }
  }

  /**
   * Simple list management that avoids synchronized/functional interface of Vector.  Key to this
   * implementation is that all changes are made to a "copy" of the original list.  This allows
   * for non-synchronized access of listners when event dispatch is called.
   */
  public int find(EventListener listener) {
    if (listeners != null ) {
      for (int index = 0; index < listeners.length; ++index)
        if (listeners[index] == listener)
          return index;
    }
    return -1;
  }


  /**
   *
   */
  public synchronized final void add(EventListener listener) {
    if (find(listener) < 0) {
      EventListener[] newListeners;

      if (listeners == null)
        newListeners = new EventListener[1];
      else {
        newListeners = new EventListener[listeners.length+1];
        System.arraycopy(listeners, 0, newListeners, 0, listeners.length);
      }

      newListeners[newListeners.length-1] = listener;
      listeners = newListeners;  // Assumed atomic.
    }
  }

  /**
   *
   */
  public synchronized final void remove(EventListener listener) {
    int index = find(listener);
    if (index > -1) {
      // Important: hasListeners() expects listeners too be null if there are no listeners.
      if (listeners.length == 1)
        listeners = null;
      else {
        EventListener[] newListeners = new EventListener[listeners.length-1];
        System.arraycopy(listeners, 0, newListeners, 0, index);

        if (index < newListeners.length)
          System.arraycopy(listeners, index+1, newListeners, index, newListeners.length-index);

        listeners = newListeners;  // Assumed atomic.
      }
    }
  }

  /**
   * Used for efficiency to avoid allocating an EventMultiCaster until (or if) there are
   * any listeners.  Also allows the caller to be slightly more efficient about whether there
   * are any listeners because the EventMultiCaster member can be tested for null to see if
   * there are any listeners.
   */
  public final static EventMulticaster add(EventMulticaster caster, EventListener listener) {
    if (caster == null)
      caster = new EventMulticaster();
    Diagnostic.check(caster.find(listener) == -1);
    caster.add(listener);
    return caster;
  }                  

  public final static EventMulticaster remove(EventMulticaster caster, EventListener listener) {
    if (caster != null) {
      caster.remove(listener);
      if (!caster.hasListeners())
        caster = null;
    }
    return caster;
  }

  public int getListenerCount() {
    return listeners != null ? listeners.length : 0;
  }

  private transient EventListener[] listeners;
}
