/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */
package borland.jbcl.util;

import java.lang.*;
import java.util.*;

/**
* This FastStringBuffer class was invented to use instead of StringBuffer in those instances when we know
* a buffer is not shared.  It gets away from some of the complications of synchronization and sharing.
*/
public final class FastStringBuffer {

  private static char[] logicalLiteral = {'\r','\n','\t','\'','\\'};
  private static char[] displayLiteral = { 'r', 'n', 't','\'','\\'};

  /**
   * FastStringBuffer.NOTACHAR is used as a return value to indicate when a fetch has run out of bounds
   */
  public static final int NOTACHAR = 0;
  public static final int NOT_A_CHAR = 0;

  private char value[];  // the array we keep things in
  private int count;    // number of elements in value[]. Note! not always same as value.length!
  private int offset;    // used to communicate current position when scanning

  public FastStringBuffer() {
    this(16);
  }

  public FastStringBuffer(int length) {
    value = new char[length];
    offset = 0;
    count = 0;
  }

  public FastStringBuffer(String str) {
    this(str.length() + 16);
    append(str);
    offset = 0;
    count = str.length();
  }

  public FastStringBuffer(char[] cArray) {
    this(cArray.length + 16);
    append(cArray);
    offset = 0;
    count = cArray.length;
  }

  public FastStringBuffer(char[] cArray, int offset, int len) {
    this(len + 16);
    append(cArray, offset, len);
    offset = 0;
    count = len;
  }

  public FastStringBuffer(char c, int nChars) {
    this(nChars);
    for (int i = 0; i < nChars; ++i)
      value[i] = c;
    count = nChars;
  }


  public void empty() {
    for (int i = 0; i < count; i++)
      value[i] = '\0';
    count = 0;
    offset = 0;
  }

  public char[] value() {
    return value;
  }

  public char[] getValue() {
    return value;
  }

  public int length() {
    return count;
  }

  public int getLength() {
    return count;
  }

  public int capacity() {
    return value.length;
  }

  public int offset() {
    return offset;
  }

  public int getOffset() {
    return offset;
  }

  public void setOffset(int offset) {
    this.offset = offset;
  }

  public void makeroom(int minimumCapacity) {
    int maxCapacity = value.length;

    if (minimumCapacity > maxCapacity) {
      int newCapacity = (maxCapacity + 1) * 2;
      if (minimumCapacity > newCapacity)
        newCapacity = minimumCapacity;
      char[] newValue = new char[newCapacity];
      System.arraycopy(value, 0, newValue, 0, count);
      value = newValue;
    }
  }

  public void setLength(int newLength) {
    if (newLength < 0)
      throw new StringIndexOutOfBoundsException(newLength);
    makeroom(newLength);

    if (count < newLength) {
      for (; count < newLength; count++) {
        value[count] = '\0';
      }
    }
    count = newLength;
  }

  public char charAt(int index) {
    if (index < 0 || index >= count)
      throw new StringIndexOutOfBoundsException(index);
    offset = index;
    return value[index];
  }

  // Returns the first char in the FastStringBuffer -- used for loops.
  public char firstChar() {
    if (count <= 0)
      return NOTACHAR;
    offset = 0;    // gets ready for nextChar()
    return value[0];
  }

  /**
   * Returns the last character in the string and positions the offset
   * there.  Meant to be used as lastChar()/priorChar() loop
   *
   * @return The final char (or FastStringBuffer.NOTACHAR) if empty)
   */
  public char lastChar() {
    if (count <= 0)
      return NOTACHAR;
    offset = count-1;
    return value[offset];
  }

  // Returns the character at value[offset] -- used for loops.  NOTE!!! this STARTS at offset.
  public char currentChar() {
    if ((offset < 0) || (offset >= count))
      return NOTACHAR;
    return value[offset];
  }

  // Returns next sequential char (value[offset+1).  Zero return means done.
  public char nextChar() {
    if (offset < -1 || offset >= count-1)
      return NOTACHAR;
    return value[++offset];
  }

  // Returns previous sequential char (value[offset-1).  Zero return means done.
  public char priorChar() {
    if (offset < 1 || offset > count)
      return NOTACHAR;
    return value[--offset];
  }

  /**
   * To be used in a firstChar()/nextChar() sort of loop, this method
   * peeks at the next character WITHOUT advancing any pointers.
   */
  public char peekNextChar() {
    if (offset < -1 || offset >= count-1)
      return NOTACHAR;
    return value[offset+1];
  }

  public void getChars(int srcBegin, int srcEnd, char dst[], int dstBegin) {
    if (srcBegin < 0 || srcBegin >= count)
      throw new StringIndexOutOfBoundsException(srcBegin);
    if (srcEnd < 0 || srcEnd > count)
      throw new StringIndexOutOfBoundsException(srcEnd);
    if (srcBegin < srcEnd)
      System.arraycopy(value, srcBegin, dst, dstBegin, srcEnd - srcBegin);
  }

  public void setCharAt(int index, char ch) {
    if (index < 0 || index >= count)
      throw new StringIndexOutOfBoundsException(index);
    value[index] = ch;
  }

  /**
   * This method removes the character from the buffer at value[offset].  It adjusts 'offset' so
   * a nextChar() loop will find the character immediately following the one removed.
   */
  public void removeCharAt(int index) {
    if (index < 0 || index >= count)
      throw new StringIndexOutOfBoundsException(index);
    int len = count - index - 1;
    if (index < count-1 && len > 0)
      System.arraycopy(value, index+1, value, index, len);
    offset = index - 1;    // so nextChar() will see next sequential char
    --count;
  }

  /**
   * This method removes the "current" character from the buffer, where "current" is defined
   * by 'offset'. It is intended to be used in a firstChar/nextChar loop.  It will adjust 'offset'
   * so that the next nextChar() will function properly.
   */
  public void removeChar() {
    removeCharAt(offset);
  }

  /**
   * This method will remove a number of characters from the buffer
   */
  public void removeCharsAt(int index, int removeCount) {
    if (index < 0 || index+removeCount >= count)
      throw new StringIndexOutOfBoundsException(index);
    if (index < count-removeCount)
      System.arraycopy(value, index+removeCount, value, index, count-index-removeCount);
    offset = index - removeCount;    // so nextChar() will see next sequential char
    count -= removeCount;
  }

  /**
   * This method removes the specified number of characters from the current position (where
   * "current" is defined as value[offset].  It adjusts 'offset' so the next nextChar() will
   * encounter the next character in the buffer.
   */
  public void removeChars(int removeCount) {
    removeCharsAt(offset, removeCount);
  }

  public void replaceCharAt(int index, char c) {
    if (index < 0 || index >= count)
      throw new StringIndexOutOfBoundsException(index);
    value[index] = c;
  }

  public FastStringBuffer append(Object obj) {
    return append(String.valueOf(obj));
  }

  public FastStringBuffer append(char c) {
    makeroom(count + 1);
    value[count++] = c;
    return this;
  }

  public FastStringBuffer append(char c, int appendCount) {
    makeroom(count + appendCount);
    for (int i = 0; i < appendCount; ++i)
      value[count++] = c;
    return this;
  }

  public FastStringBuffer append(String str) {
    if (str == null)
      str = String.valueOf(str);

    int len = str.length();
    makeroom(count + len);

    str.getChars(0, len, value, count);
    count += len;
    return this;
  }

  public FastStringBuffer append(FastStringBuffer fsb) {
    if (fsb == null)
      return this;

    int len = fsb.count;
    makeroom(count + len);

    fsb.getChars(0, len, value, count);
    count += len;
    return this;
  }

  public FastStringBuffer append(char[] str) {
    int len = str.length;
    makeroom(count + len);

    System.arraycopy(str, 0, value, count, len);
    count += len;
    return this;
  }

  public FastStringBuffer append(char[] str, int offset, int len) {
    makeroom(count + len);

    System.arraycopy(str, offset, value, count, len);
    count += len;
    return this;
  }

  public FastStringBuffer insert(int offset, Object obj) {
    return insert(offset, String.valueOf(obj));
  }

  public FastStringBuffer insert(int offset, String str) {
    if (offset < 0 || offset > count)
      throw new StringIndexOutOfBoundsException();
    int len = str.length();
    makeroom(count + len);

    System.arraycopy(value, offset, value, offset + len, count - offset);
    str.getChars(0, len, value, offset);
    count += len;
    return this;
  }

  public FastStringBuffer insert(int offset, char[] str) {
    if (offset < 0 || offset > count)
      throw new StringIndexOutOfBoundsException();
    int len = str.length;
    makeroom(count + len);

    System.arraycopy(value, offset, value, offset + len, count - offset);
    System.arraycopy(str, 0, value, offset, len);
    count += len;
    return this;
  }

  public FastStringBuffer insert(int offset, boolean b) {
    return insert(offset, String.valueOf(b));
  }

  public FastStringBuffer insert(int offset, char c) {
    makeroom(count + 1);

    System.arraycopy(value, offset, value, offset + 1, count - offset);
    value[offset] = c;
    count += 1;
    return this;
  }

  // Finds the index in the given FastStringBuffer where the specified subString
  // begins.  -1 if not found.
  //
  public int IndexOfSubString(FastStringBuffer subStr, int fromIndex) {
    return indexOf(subStr, fromIndex);
  }

  /**
   * Just like String.indexOf but uses FastStringBuffer
   */
  public int indexOf(FastStringBuffer subStr, int fromIndex) {
    char[] v1 = value;
    char[] v2 = subStr.value();
    int max = count - subStr.length();

    test:
    for (int i = ((fromIndex < 0) ? 0 : fromIndex); i <= max ; i++) {
      int n = subStr.length();
      int j = i;
      int k = 0;
      while (n-- != 0) {
        if (v1[j++] != v2[k++]) {
          continue test;
        }
      }
      return i;
    }
    return -1;
  }

  /**
   * Just like String.lastIndexOf but uses FastStringBuffer
   */
  public int lastIndexOf(FastStringBuffer subStr, int fromIndex) {
    char[] v1 = value;
    char[] v2 = subStr.value();
    int max = count - subStr.length();

    test:
    for (int i = fromIndex; i >= 0 ; --i) {
      int n = subStr.length();
      int j = i;
      int k = 0;
      while (n-- != 0) {
        if (v1[j++] != v2[k++]) {
          continue test;
        }
      }
      return i;
    }
    return -1;
  }

  /**
   * Just like String.substring() but uses FastStringBuffer
   */
  public FastStringBuffer substring(int startPos, int endPos) {
    FastStringBuffer fsb;
    if (startPos < 0)
      startPos = 0;
    if (endPos >= count)
      endPos = (count - 1);
    if (startPos >= endPos)
      fsb = new FastStringBuffer();
    else fsb = new FastStringBuffer(value, startPos, endPos-startPos);
//System.err.println("fsb.subString(" + this + ", " + startPos + ", " + endPos + ") yielded:\r\n" + fsb);
    return fsb;
  }
  
  public String toString() {
    return new String(this.value, 0, count);
  }

  /**
   * Given a FastStringBuffer where charAt(), nextChar() or priorChar()
   * have just returned the backslash character (in other words, where
   * value[offset] == '\\', this routine will parse the rest as a single
   * character backslash value (e.g. "\u2003"), advancing offset.  It
   * will return that character and leave the FastStringBuffer pointing
   * at the next character after the value.
   *
   * @return The single character value.
   */
  public char parseBackSlash() {
    char c = nextChar();
    char cVal = displayToLiteral(c);  // if 'r', 'n', 't', etc, return '\r', '\n', or '\t', respectively
    if (cVal != NOTACHAR)
      return cVal;

    int radix;
    int maxChars;
    switch (c) {
      case 'U':
      case 'u':
        radix = 16;
        maxChars = 4;
        break;
      case '0':
        radix = 8;
        maxChars = 3;
        c = peekNextChar();
        if (c == 'x' || c == 'X') {
          radix = 16;
          maxChars = 2;
          ++offset;
        }
        else priorChar();
        break;

      default:
        return c;
    }

    FastStringBuffer s = new FastStringBuffer(maxChars);
    for (int i = 0; i < maxChars; ++i) {
      c = Character.toUpperCase(nextChar());
      if (!Character.isDigit(c)) {
        if (radix == 8 && c > '7' || radix == 16 && !(c >= 'A' &&  c <= 'F')) {
          priorChar();
          break;
        }
      }
      s.append(c);
    }
    return (char) Integer.parseInt(s.toString(), radix);
  }

  /**
   * Given a string which needs to be parsed as a literal string (including
   * backslash characters), and assuming value[offset] is currently pointing
   * at the starting delimiter of this string, this routine will buffer
   * everything up to (but not including) another delimiter like the first.
   * It will advance offset past that delimiter so subsequent string
   * processing can continue.
   *
   * @param delimiter the char value which will mark the end of the literal
   *
   * @param allowDouble true means two delimiters in a row evaluate to a single
   * occurance of that literal in the string (and it is not a delimiter)
   *
   * @return It returns a new FastStringBuffer containing the literal.
   */
  public FastStringBuffer parseLiteral(char delimiter, boolean allowDouble) {
    FastStringBuffer rsb = new FastStringBuffer();
    for (char c = nextChar(); c != NOTACHAR; c = nextChar()) {

      // Delimiter encountered -- it allowDouble and see 2 in a row,
      // then treat it as an embedded literal and keep scanning
      if (c == delimiter) {
        if (allowDouble && peekNextChar() == c) {
          rsb.append(c);
          nextChar();   // skip 2nd delimiter
          continue;
        }
        break;
      }

      // A backslash gets special evaluation
      if (c == '\\')
        c = parseBackSlash();
      
      rsb.append(c);
    }
    return rsb;
  }

  /**
   * Given a string which needs to be parsed as a literal string (including
   * backslash characters), and assuming value[offset] is currently pointing
   * at the starting delimiter of this string, this routine will buffer
   * everything up to (but not including) another delimiter like the first.
   * It will advance offset past that delimiter so subsequent string
   * processing can continue.
   *
   * @return It returns a new FastStringBuffer containing the literal.
   */
  public FastStringBuffer parseLiteral() {
    return parseLiteral(currentChar(), false);
  }

  /**
   * Creates a copy of the given FastStringBuffer, but will translate any characters in the
   * specified delimiter set into unicode "escape" sequences (e.g. ',' becomes '\u002C').  This will allow
   * the new FastStringBuffer to use the normal StringTokenizer for parsing.
   *
   * @param sourceString Contains a String to be scanned and converted (though this
   * String itself will not be altered).
   *
   * @param delimiters Contains a String consisting of the delimiters you don't want to see
   * in the output StringBuffer (e.g. new String("\t\r\n,").  Wherever these occur in the
   * sourceString, they will be converted to a unicode escape sequence.
   *
   * @return A new FastStringBuffer containing all the characters of sourceString, but with
   * all delimiters expanded to unicode escape sequences.
   */
  public static FastStringBuffer expandDelimiters(String sourceString, String delimiters) {

    int len = (sourceString == null) ? 0 : sourceString.length();
    FastStringBuffer result = new FastStringBuffer(len);
    for (int i = 0; i < len; ++i) {
      char c = sourceString.charAt(i);
      if (delimiters.indexOf(c) >= 0)
        result.append(charToUnicodeEscape(c));
      else
        result.append(c);
    }
    return result;
  }

  /**
   * The opposite of expandDelimiters, this method will turn any unicode escape sequences that
   * would result in one of the given delimiters characters into the displayable form of that
   * delimiters. Do NOT pass in a nondisplayable delimiter (like '\r')
   *
   */
  public FastStringBuffer normalizeDelimiters(String delimiters) {
    FastStringBuffer result = new FastStringBuffer(count);
    int oldoffset;
    for (char c = firstChar(); c != NOTACHAR; c = nextChar()) {
      if (c == '\\') {
        oldoffset = offset;
        char cVal = parseBackSlash();    // get char literal following backslash
        if (delimiters.indexOf(cVal) >= 0) {    // Is it in our delimiter list?
          char cLog = literalToDisplay(cVal);    // Yes, can we convert to logical? (e.g. '\r', etc.)
          if (cLog != NOTACHAR) {
            result.append('\\');      // Yes, form readable version of literal
            result.append(cLog);
            continue;
          }
          result.append(cVal);      // No readable form, just append literal itself
          continue;
        }
        offset = oldoffset;        // unicode sequence not a delimiter -- leave as raw unicode
      }
      result.append(c);
    }
    return result;
  }

  /**
   * Returns a String containing a unicode escape sequence representing the given character.
   * For example charToUnicodeEscape('1') would return new String("\\u0031").
   *
   * @param ch The character to be converted to a unicode escape sequence
   *
   * @return A new String containing the escape sequence
   */
  public static String charToUnicodeEscape(char ch) {
    char hexChars[] = new char[6];
    hexChars[0] = '\\';
    hexChars[1] = 'u';
    hexChars[2] = Hex.chars[(ch >> 12) & 0xF];
    hexChars[3] = Hex.chars[(ch >>  8) & 0xF];
    hexChars[4] = Hex.chars[(ch >>  4) & 0xF];
    hexChars[5] = Hex.chars[ch & 0xF];
    return new String(hexChars, 0, 6);
  }

  private static char displayToLiteral(char c) {
    for (int i = 0; i < displayLiteral.length; ++i)
      if (c == displayLiteral[i])
        return logicalLiteral[i];
    return NOTACHAR;
  }

  private static char literalToDisplay(char c) {
    for (int i = 0; i < logicalLiteral.length; ++i)
      if (c == logicalLiteral[i])
        return displayLiteral[i];
    return NOTACHAR;
  }

  /**
   * This method returns the first "logical" char value from the given
   * String.  This means it handles backslashes, unicode escape sequences, etc.
   */
  public static char charFromString(String s) {
    //System.err.println("charFromString(" + s + ")");
    char c = (char) Character.UNASSIGNED;
    if (s != null && s.length() >= 0) {
      if (s.length() == 1)
        c = s.charAt(0);
      else {
        FastStringBuffer fsb = new FastStringBuffer(s);
        c = fsb.firstChar();
        if (c == '\\')
          c = fsb.parseBackSlash();
      }
    }
    //System.err.println(" returning <" + c + "> which is value: " + (int) c);
    
    return c;
  }


  /**
   * The opposite of charFromString, this method will return a String which best
   * represents the given char.  This means it expands it into a unicode escape
   * sequence if needed.
   */
  public static String stringFromChar(char c) {
    String s;
    //System.err.println("stringFromChar(" + (int) c + ")");
    char cLog = literalToDisplay(c);
    if (cLog != NOTACHAR)  
      s = "\\" + cLog;
    // Anything less than SPACE or 2-bytes goes to unicode
    else if ((int) c < 32 /*|| (int) c > 256*/)
      s = charToUnicodeEscape(c);
    else
      s = new String(new char[] {c});

    //System.err.println(" returning <" + s + ">");
    return s;
  } 

  /**
   * Translates a String which is compatible with source code (including
   * leading and trailing quote, expands backslash charaters, etc.) into
   * its actual String representation (e.g. the literal "\n" becomes the
   * real linefeed character, etc.)
   */
  public static FastStringBuffer sourceToText(String source) {
//System.err.println("sourceToText(" + source + ")");    
    FastStringBuffer text = new FastStringBuffer();
    FastStringBuffer input = new FastStringBuffer(source);

    // Strip optional leading and trailing quotes
    if (input.charAt(0) == '\"')
      input.removeCharAt(0);
    int len = input.length();
    if (len > 0 && input.charAt(len-1) == '\"')
      input.removeCharAt(len-1);
      
    for (char ch = input.firstChar(); ch != NOTACHAR; ch = input.nextChar()) {
      //System.err.println("sourceToText is " + ch);
      if (ch == '\\')
        ch = input.parseBackSlash();
//System.err.println("appending char <" + (int) ch + "> which is [" + ch + "]");
      //System.err.println("  appending char val " + (int) ch);
      text.append(ch);
    }
    return text;
  }
  
  static final String needingEscape = "\b\t\n\f\r\"\\";  //NORES

  /**
   * Converts a String into a form that will compile (translating
   * special characters into backslash-something and adding
   * leading and trailing quotes.
   */
  public static FastStringBuffer textToSource(String text, boolean hasEscapes) {
    return textToSource(text, hasEscapes, "    ");
  }

  private static final int SUGGESTED_MAX_SOURCE_COLS = 64;    // max columns before line break (if can find whitespace)
  private static final int ABSOLUTE_MAX_SOURCE_COLS = 100;    // max columns before line break -- will NOT exceed this
  
  public static FastStringBuffer textToSource(String text, boolean hasEscapes, String indentString) {
    FastStringBuffer source = new FastStringBuffer(text.length()*2+2);
    source.append("\"");
    //System.err.println(" length=" + text.length() + " hasEscapes=" + hasEscapes);
    int colCount = 0;
    
    for (int i = 0; i < text.length(); i++) {
      char ch = text.charAt(i);
      //System.err.println("char[" + i + "]=" + (int) ch );
      
      // If we are anticipating escape sequences, look ahead to see
      // whether a backslash is followed by one of the legal escape
      // prefixes and then by a digit -- that's enough to presume
      // the backslash does not get a 2nd backslash
      if (ch == '\\' && hasEscapes && i+2 < text.length()) {
        //System.err.println(" lookahead = <" + text.charAt(i+1) + ">  and  <" + text.charAt(i+2) + ">");
        char digit = text.charAt(i+2);
        switch (text.charAt(i+1))
        {
          case '0':
            if ((digit == 'x' || digit == 'X') && i+3 < text.length())
              digit = text.charAt(i+3);
            // falls into
          case 'u':
          case 'U':
            if (digit < '0' || digit > '9')
              break;
            // falls into
            //System.err.println("  looks like backslash escape");
            source.append(ch);
            continue;
        }
      }
      //System.err.println("  not escape");
      if (needingEscape.indexOf(ch) >= 0) {
        //System.err.println("  needing escape");
        source.append("\\");
        switch (ch) {
          case '\b': source.append('b'); break;
          case '\t': source.append('t'); break;
          case '\n': source.append('n'); break;
          case '\f': source.append('f'); break;
          case '\r': source.append('r'); break;

          // 2 backslashes in a row echo through as 2 backslashes
          case '\\': source.append('\\');
                     if (hasEscapes && (i+1) < text.length() && text.charAt(i+1) == '\\')
                       source.append(text.charAt(++i));
                     break;
                     
          default:  source.append(FastStringBuffer.stringFromChar(ch)); break;
        }
      }
      else
        source.append(FastStringBuffer.stringFromChar(ch));

      // Because the compiler can accept lines no longer than 1024, and because
      // really long lines look ugly, we try to break them when then exceed a
      // low threshhold if we see whitespace.  But in no case do we exceed the
      // absolute limit on length without breaking
      ++colCount;
      if (i < (text.length() - 1) && colCount > SUGGESTED_MAX_SOURCE_COLS) {
        boolean breakHere = Character.isWhitespace(ch);
        if ((breakHere && colCount > SUGGESTED_MAX_SOURCE_COLS) ||
            colCount > ABSOLUTE_MAX_SOURCE_COLS) {
          source.append("\" +\n");
          source.append(indentString);
          source.append("\"");
          colCount = 0;
        }
      }
    }
    source.append("\"");
    return source;
  }

}
