/*
 * Copyright (c) 1997-1998 Borland International, Inc. All Rights Reserved.
 * 
 * This SOURCE CODE FILE, which has been provided by Borland as part
 * of a Borland product for use ONLY by licensed users of the product,
 * includes CONFIDENTIAL and PROPRIETARY information of Borland.  
 *
 * USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS 
 * OF THE LICENSE STATEMENT AND LIMITED WARRANTY FURNISHED WITH
 * THE PRODUCT.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD BORLAND, ITS RELATED
 * COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST ANY CLAIMS
 * OR LIABILITIES ARISING OUT OF THE USE, REPRODUCTION, OR DISTRIBUTION
 * OF YOUR PROGRAMS, INCLUDING ANY CLAIMS OR LIABILITIES ARISING OUT OF
 * OR RESULTING FROM THE USE, MODIFICATION, OR DISTRIBUTION OF PROGRAMS
 * OR FILES CREATED FROM, BASED ON, AND/OR DERIVED FROM THIS SOURCE
 * CODE FILE.
 */

package borland.jbcl.util;

import java.lang.*;
import java.util.*;
import java.text.*;

/**
 * This method extends the basic Format class but allows for special formatting of text.  Like the other
 * Format derivatives, it is based on a control pattern.  The format of the control pattern is as follows:
 * 
 *     pattern ; keepLiterals ; fillCharacter ; replaceCharacter
 * 
 * @param pattern can contain any of the following characters (they are a common format shared by
 * many products):
 * 
 *     0		Digit 0:9, entry required, '+' and '-' not allowed
 *     9		Digit 0:9, entry optional, '+' and '-' not allowed
 *     #		Digit or space, entry optional, plus and minus signs allowed
 *     L		Letter A:Z, entry required
 *     l		Letter A:Z, entry optional
 *     ?		Letter A:Z, entry optional
 *     A		Letter A:Z or digit 0:9, entry required
 *     a		Letter A:Z or digit 0:9, entry optional
 *     C		Any character or space, entry required
 *     c		Any character or space, entry optional
 *     &		Any character or space, entry required
 *     <		Causes all characters following to be converted to lowercase
 *     >		Causes all characters following to be converted to uppercase
 *     !		Causes strings too short to fill from right to left
 *     \		Backslash escape -- allows any Unicode value to follow (e.g. "\u2003")
 *     ^		Initial place for cursor when editing begins
 *     ''		Encloses a literal expression (for example, the pattern
 * 		"990' units sold'" would display as "27 units sold")
 *     **		Encloses a password encrypted string (for example,
 * 		the pattern "*AAAAaaaa*" would accept a password at
 * 		least 4, and at most 8 alphanumeric characters).  
 * 		Characters typed or displayed appear as the '*' character
 *     
 *     An example of a US phone number might be "!(999)000-0000"
 * 
 * @param keepLiterals If this value is "0", then literals in the string will be removed before
 * it is stored.  For example (408)555-5330 would become 4085555330.  Any other value (including
 * no value at all) will default to "1" which means the literals are preserved.
 * 
 * @param fillCharacter Whenever a string is formatted which is too short to fill all the available
 * positions in the pattern, this character will be used to fill the extra space.  For example,
 * formatting "(   )555-5330 with the pattern "!(999)000-0000;1;* will produce "(***)555-5330".
 * Note that a value of zero (communicated via "\0" means no filling will occur.
 * 
 * @param replaceCharacter Whenever a string is parsed, all occurances of "fillCharacter" are replaced
 * with "replaceCharacter".  Using the example above, parsing (***)555-5330 with the pattern
 * "!(999)000-0000;1;*;_" will produce (___)555-5330.  Again, a zero value is allowed in this field
 * (using "\0" and will have the effect of removing the fillCharacters.
 * 
 * <P><B>Note:</B> Each of the special fields (keepLiterals, fillCharacter, replaceCharacter) have their
 * own methods to get or set them.    
 */
public class TextFormat extends java.text.Format {
  
  public static char NOT_A_CHAR = 0xffff;
//  static char PasswordChar = 0xfffe;
  static final String TRACE = "TextFormat";

  static final byte LITERAL_CHAR =  (byte) 0x01;
  static final byte OPTIONAL_CHAR = (byte) 0x02;
  static final byte PASSWORD_CHAR = (byte) 0x04;
  
  private String pattern;
  boolean rightToLeft;
  boolean injectLiterals;
  char[] litMask;
  byte[] infoMask;
  int charCount;		// true count of characters we can parse/format
  int litCount;			// count of literal chars which will be seen/injected
  char fillChar;		// the char which is interpreted to mean "this is not a char"
  char replaceChar;			// the char that will replace fillChar (0 = strip fillChars)
  boolean explicitFillChar;
  
  public TextFormat () {
    applyPattern(null);
    //Diagnostic.addTraceCategory(TRACE);
}
  
  public TextFormat (String pattern) {
    this();
    applyPattern(pattern);
  }
  
  
  public void setFillCharacter(char c) {
    if (c == NOT_A_CHAR) {
      fillChar = ' ';
      explicitFillChar = false;
    }
    else  {
      fillChar = c;
      explicitFillChar = true;
    }
  }
  
  public char getFillCharacter() {
    return (explicitFillChar) ? fillChar : NOT_A_CHAR;
  }
  
  public void setReplaceCharacter(char c) {
    replaceChar = c;
  }
  
  public char getReplaceCharacter() {
    return replaceChar;
  }
  
  public void setKeepLiterals(boolean tf) {
    injectLiterals = !tf;
  }
  
  public boolean getKeepLiterals() {
    return !injectLiterals;
  }
  
  public void applyPattern(String pattern) {
    this.pattern = (pattern != null) ? new String(pattern) : null;
    decomposePattern();
  }
  
  public String toPattern() {
    return (pattern == null) ? null : pattern.toString();
  }
  
  private void decomposePattern() {
    
    charCount = 0;		// a null pattern still needs to init some stuff
    litCount = 0;
    rightToLeft = false;
    injectLiterals = false;
    fillChar = ' ';		// character which means "this is not a real char"
    explicitFillChar = false;
    replaceChar = 0;		// char to replace fillChar with (0 = strip out fillChar's)
    litMask = null;
    if (pattern == null)
      return;
    
    Diagnostic.trace(TRACE, "Decomposing pattern: " + pattern);
    //new NullPointerException().printStackTrace();
    
    int subRegion = 0;
    int iLen = pattern.length();
    FastStringBuffer fsb = new FastStringBuffer(pattern);
    FastStringBuffer fsb2;
    int i;
    char c;
    int passwordCount = 0;
    //
    // The first pass collects information about the string (e.g. size, literals, etc.)
    //
    for (c = fsb.firstChar(); c != FastStringBuffer.NOTACHAR; c = fsb.nextChar()) {
      switch (c)
      {
        //
        // The ';' char is a region separator.  They are controlString;stripLiterals;fillChar;replaceChar
        // Any region (but the first) maybe omitted, and the defaults will be used.
        // For example "!(999)000-0000;0;;*" would have a default fillChar and '*' as a replaceChar
        //
        case ';':
          ++subRegion;
          if (fsb.offset() > iLen-1)	    // beware end of string
            c = 0;
          else c = fsb.peekNextChar();    // 2 semi's mean nothing here
          Diagnostic.trace(TRACE, "   next char is \'" + c + "\'");
          if (c == ';')
            c = 0;
          //else if (c == '\\') {	    // could be backslash constant
          //  c = fsb.parseBackSlash();
          //  Diagnostic.trace(TRACE, "backslash at region " + subRegion + " is " + (int) c);
          // }
          else c = fsb.nextChar();	    // else take next character
          if (c == '\\') {
            c = fsb.parseBackSlash();
            Diagnostic.trace(TRACE, "backslash at region " + subRegion + " is " + (int) c);
          }
          switch (subRegion)
          {
            case 1:
              if (c == '0')
                injectLiterals = true;
              continue;
            case 2:
              fillChar = c;
              explicitFillChar = true;
              continue;
            case 3:
              replaceChar = c;
              //System.err.println("TextFormat sees replaceChar as <" + c + ">");              
              // falls into
            default:
              fsb.lastChar();	// will stop the loop above
          }
          continue;
          //
          // We're just counting literal and character positions in this pass
          //
          
        case '!':
          rightToLeft = true;
          // falls into
        case '>':
        case '<':
        case '^':
        case '{':
        case '}':
          continue;
        case '*':	// ** brackets passwords
        case '\'':	// single quotes bracket a literal
          fsb2 = fsb.parseLiteral(c, (c == '\''));    // allow 2 quotes to become single quote
          if (c == '*') {
            ++passwordCount;
            charCount += fsb2.length();
          }
          else litCount += fsb2.length();
          continue;
        case '\\':
          fsb.parseBackSlash();
          ++litCount;
          continue;
          
        case '0':
        case '9':
        case '#':
        case 'L':
        case 'l':
        case '?':
        case 'A':
        case 'a':
        case '&':
        case 'C':
        case 'c':
          ++charCount;
          continue;
          
        default:	// everything else is a literal
          ++litCount;
      }
    }
    
    
    if (injectLiterals)
      Diagnostic.trace(TRACE, "We are injecting/stripping literals");
    else
      Diagnostic.trace(TRACE, "We are injecting " + litCount + " literals");
    if (rightToLeft)
      Diagnostic.trace(TRACE, "We are scanning right to left");
    Diagnostic.trace(TRACE, "The fillChar is \'" + fillChar + "\'");
    Diagnostic.trace(TRACE, "The replaceChar is \'" + replaceChar + "\'");
    
    //
    // We now know how many characters belong in the string and how many literals were encountered.
    // If we don't have to strip/inject literals, then we have a very simple case (where we only reproduce
    // the string as given).
    //
    
    if (litCount == 0 && passwordCount == 0) {
      Diagnostic.trace(TRACE, "Pattern has no literals or password chars");
      return;
    }
    
    //
    // Let's construct a map for every character we would see in this display.
    // Every place there is a fillable character spot is marked "NOT_A_CHAR",
    // every place that should echo as a password "*" is marked PasswordChar,
    // and every literal part of the string comes straight from that literal.
    Diagnostic.trace(TRACE, "Pattern has " + litCount + " literals and " + charCount + " char slots");
    
    int fillPos = 0;
    litMask = new char[charCount+litCount];
    infoMask = new byte[charCount+litCount];
    
    for (i = 0; i < litMask.length; ++i) {
      litMask[i] = NOT_A_CHAR;
      infoMask[i] = 0;
    }

    byte infoBits = 0;

    for (c = fsb.firstChar(); c != FastStringBuffer.NOTACHAR; c = fsb.nextChar()) { 
      Diagnostic.trace(TRACE, "considering char " + c);
      switch(c)
      {
        case '!':	// directives occupy no room
        case '>':
        case '<':
        case '^':
          continue;
        case '{':
          infoBits |= (OPTIONAL_CHAR);
          //Diagnostic.trace(TRACE, "infoBits = " + infoBits);
          continue;
          
        case '}':
          infoBits &= ~(OPTIONAL_CHAR);
          //Diagnostic.trace(TRACE, "infoBits = " + infoBits);
          continue;
          
        case '*':	// password shares parsing logic
          infoBits ^= (PASSWORD_CHAR);
          continue;
          
        case '\'':	// single quotes bracket a literal
          fsb2 = fsb.parseLiteral(c, (c == '\''));    // double quotes means single quote literal
          System.arraycopy(fsb2.value(), 0, litMask, fillPos, fsb2.length());
          for (int j = 0; j < fsb2.length(); ++j)
            infoMask[fillPos++] = LITERAL_CHAR;
          //fillPos += fsb2.length();
          continue;

        case '\\':
          infoMask[fillPos] = LITERAL_CHAR;
          litMask[fillPos++] = fsb.parseBackSlash();
          continue;
          
        case '0':
        case 'L':
        case 'A':
        case 'c':
          //Diagnostic.trace(TRACE, " infoBits for '" + c + "' at pos " + fillPos + " = " + infoBits);
          infoMask[fillPos++] = infoBits;
          continue;
          
        case '9':
        case '#':
        case 'l':
        case '?':
        case 'a':
        case '&':
        case 'C':
          infoMask[fillPos++] = (byte) ((infoBits & PASSWORD_CHAR) | OPTIONAL_CHAR);
          continue;
          
        case ';':
          fsb.lastChar();	// to stop the loop
          break;
          
        default:	// everything else is a literal
          Diagnostic.trace(TRACE, "Litmask[" + fillPos + "] gets the char \'" + c + "\'");
          infoMask[fillPos] = LITERAL_CHAR;
          litMask[fillPos++] = c;
          continue;
      }
    }
    
    Diagnostic.trace(TRACE, "LitMask is " + litMask.length + " and we filled " + fillPos);
    


    if (Diagnosing.on) {
      StringBuffer dStr = new StringBuffer();
      for (i = 0; i < litMask.length; ++ i)
        dStr.append((litMask[i] == NOT_A_CHAR)
                    ? '_'
                    : ((infoMask[i] & PASSWORD_CHAR) != 0)
                      ? '*'
                      : litMask[i]);
      System.err.println(dStr.toString() + "  <-- literals");

      dStr = new StringBuffer();
      for (i = 0; i < infoMask.length; ++ i)
        dStr.append(((infoMask[i] & LITERAL_CHAR) != 0)
                    ? litMask[i]
                    : ((infoMask[i] & OPTIONAL_CHAR) != 0)
                      ? '_'
                      : '!');
      System.err.println(dStr.toString() + "  <-- !=required");
    }
  }
  
  
  /**
  This method will format the given String using the pattern associated with this object.
  If the input string has insufficient characters to fill the pattern, they will be filled
  with the 'fillCharacter' character (which can be set via setFillChar()).
  
  @param toBeFormatted The input String which requires formatting
  @param result The output StringBuffer containing the formatted result
  @param status Currently not used
  
  @return The 'result' parameter is also the returned value.
  
*/
  public StringBuffer format(String toBeFormatted, StringBuffer result, FieldPosition pos) {
    Diagnostic.trace(TRACE, "format(" + toBeFormatted + ", " + result + ", " + pos + ")");
    int maxChars = charCount + litCount;  // compute size of formatting pattern
    // Quick out if there is no pattern
    if (maxChars == 0) {
      result = new StringBuffer(toBeFormatted);
      return result;
    }
    
    char[] buffer;
    int iLen = toBeFormatted.length();
    
    Diagnostic.trace(TRACE, "Formatting \"" + toBeFormatted + "\" with \"" + pattern + "\""); //$$$

    /*
    if (Diagnosing.on && litMask != null) {
      for (int j = 0; j < litMask.length; ++j) {
        if (litMask[j] == NOT_A_CHAR)
          System.err.print("_");
        else System.err.print(litMask[j]);
      }
      System.err.println("");
    }
    */
    
    Diagnostic.trace(TRACE, "iLen = " + iLen + ", maxChars = " + maxChars + ", charCount = "
                       + charCount + ", litCount = " + litCount);
    /* We no longer check since it is permissible to pass in too short a formatting string
     *      if (iLen > maxChars) {
     *        throw new StringIndexOutOfBoundsException(0);
     *      }
     */
    buffer = new char[iLen];
    toBeFormatted.getChars(0, iLen, buffer, 0);
    result.setLength(0);
    
    
    int fillPos;
    int incr = 1;
    int i = 0;
    int srcPos = 0;
    int nRemaining = iLen;
    
    if (rightToLeft) {
      incr = -1;
      srcPos = iLen - 1;
      i = maxChars - 1;
    }
    
    char[] fillBuffer = new char[maxChars];
    Diagnostic.trace(TRACE, "Input buffer is " + buffer.length + ", FillBuffer is " + fillBuffer.length);
    if (litMask != null)
      Diagnostic.trace(TRACE, "litMask is " + litMask.length);
    
    for (fillPos = i; i >= 0 && i < maxChars; i += incr) {
      Diagnostic.trace(TRACE, "FillPos=" + fillPos + ", i=" + i + ", nRemaining=" +nRemaining);  //$$$

/*
      if (Diagnosing.on) {
        if (litMask != null) {
          if (litMask[i] != NOT_A_CHAR)
            Diagnostic.trace(TRACE, "LitMask[" + i + "] == " + litMask[i]);
          else Diagnostic.trace(TRACE, "LitMask expects a char here");
         }
      }
*/      
      //
      // Take care, when we run out of input chars, we continue to fill with literals and/or fillChar
      //
      if (nRemaining <= 0) {
        //if (!spotForCharacter(i)) {
        if (isLiteral(i)) {
          Diagnostic.trace(TRACE, "nRemaining <= 0, filling with litMask " + i  + " at pos " + fillPos + ": " + litMask[i]); //$$$
          Diagnostic.trace(TRACE, " pads with literal \'" + litMask[i] + "\'");
          fillBuffer[fillPos] = litMask[i];
          fillPos += incr;
          continue;
        }

        if (replaceChar != 0) {
          Diagnostic.trace(TRACE, "nRemaining <= 0, filling with replaceChar at pos " + fillPos); //$$$
          fillBuffer[fillPos] = replaceChar;
          fillPos += incr;
        }

//        else if (explicitFillChar /*&& fillChar != 0*/) {
//          Diagnostic.trace(TRACE, "nRemaining <= 0, filling with fillChar at pos " + fillPos); //$$$
//          if (fillChar != 0) {
//            fillBuffer[fillPos] = fillChar;
//            fillPos += incr;
//          }
//        }
        else Diagnostic.trace(TRACE, " being skipped");  //$$$
        continue;
      }
      //
      // Still have input chars -- it we have a literal in this position and are injecting
      // them, then inject it (but DON'T advance the input buffer pointer)
      //
      //if (!spotForCharacter(i) && ((infoMask[i] & PASSWORD_CHAR) == 0)) {
      if (isLiteral(i)) {
        Diagnostic.trace(TRACE, "literal <" + litMask[i] + "> expected at this position");
        if (injectLiterals) {
          Diagnostic.trace(TRACE, " [" + fillPos + "] gets the literal \'" + litMask[i] + "\'");
          fillBuffer[fillPos] = litMask[i];
          fillPos += incr;
        }
        //
        // If we are NOT injecting literals, perform a quick test to see whether the input text
        // has the right literal here.  It should.  If it does not, it means it was formed with
        // a different pattern.  Try injecting the literal anyway and NOT advancing the source
        // pointer.  Perhaps we will synch up eventually.  In any case, they are asking us to
        // format something which does not fit the mask, so the best we can do is try to help.
        else {
          Diagnostic.trace(TRACE, "Considering pos " + srcPos + ": " + buffer[srcPos]);  //$$$
          if (buffer[srcPos] != litMask[i] && ((infoMask[i] & PASSWORD_CHAR) == 0)) {
            Diagnostic.trace(TRACE, "Yikes: expected to see literal /'" + litMask[i] + "\' here");
            Diagnostic.trace(TRACE, "  but saw the char '\'" + buffer[srcPos] + " '\'");
            Diagnostic.trace(TRACE, "  injecting literal anyway to try to recover");
            Diagnostic.trace(TRACE, "  pos " + srcPos + " != literal: filling with replaceChar");
            if (replaceChar != 0) {
              fillBuffer[fillPos] = replaceChar;  //litMask[i]; TODO <ron> 2/19/97 bug 3293
              fillPos += incr;
            }
            continue;
          }
          else {
            if (isPassword(i)) {
              Diagnostic.trace(TRACE, " [" + fillPos + "] gets password fill");
              fillBuffer[fillPos] = '*';
            }
            else {
              Diagnostic.trace(TRACE, " [" + fillPos + "] copies \'" + buffer[srcPos] + "\' from input text");
              fillBuffer[fillPos] = buffer[srcPos];
            }
            fillPos += incr;
            srcPos += incr;	
            --nRemaining;
          }
        }
      }
      //
      // Here if we have to fill a char from the input buffer
      //
      else {
        boolean spotForChar = spotForCharacter(i);
        if (spotForChar && buffer[srcPos] == fillChar) {
          Diagnostic.trace(TRACE, " there is a fill char at this position -- injecting replaceChar");
          if (replaceChar != 0) {
            fillBuffer[fillPos] = replaceChar;
            fillPos += incr;
          }
          else {
            Diagnostic.trace(TRACE, " skipping this pos because replaceChar is nil");
          }
          srcPos += incr;
          --nRemaining;
          continue;
        }

        // If we are injecting a char at this position, look ahead and see if this char
        // is the next literal is this character.  If that is the case, we inject fillChars
        // until we synch up with that literal
        char nextLit = 0;
        if (spotForChar && litMask != null)
          for (int jj = i+incr; jj >= 0 && jj < maxChars; jj += incr)
            if (litMask[jj] != NOT_A_CHAR && ((infoMask[jj] & PASSWORD_CHAR) == 0)) {
              nextLit = litMask[jj];
              break;
            }

        Diagnostic.trace(TRACE, " (lookahead lit = \'" + nextLit + "\' <" + (int) nextLit + ">)"); //$$$

        if (nextLit != 0 && buffer[srcPos] == nextLit) {

          if (replaceChar != 0) {
            Diagnostic.trace(TRACE, "  space for char is lookahead lit -- injecting replaceChar");
            fillBuffer[fillPos] = replaceChar;
            fillPos += incr;
          }
//          else if (explicitFillChar /*&& fillChar != 0*/) {
//            Diagnostic.trace(TRACE, "  space for char is lookahead lit -- injecting fillChar");            
//            if (fillChar != 0) {
//              fillBuffer[fillPos] = fillChar;
//              fillPos += incr;
//            }
//          }
          // If there is no fillChar or replaceChar -- don't fill with anything
          continue;
        }
          
        if (litMask != null && ((infoMask[i] & PASSWORD_CHAR) != 0)) {
          Diagnostic.trace(TRACE, " [" + fillPos + "] gets password fill");
          fillBuffer[fillPos] = '*';
        }
        else {
          Diagnostic.trace(TRACE, " [" + fillPos + "] copies '" + buffer[srcPos] + "' from text");
          fillBuffer[fillPos] = buffer[srcPos];
          srcPos += incr;
        }
        fillPos += incr;
        --nRemaining;
      }
      
    }
    
    i = 0;
    if (rightToLeft)
      i = fillPos;
    if (i < 0)
      i = 0;
    if (i >= maxChars)
      i = maxChars - 1;
    
    if (rightToLeft) {
      i = fillPos+1;
      iLen = maxChars - fillPos - 1;
    }
    else {
      i = 0;
      iLen = fillPos;
    }

/*
    if (Diagnosing.on) {
      String sb = new String(fillBuffer, 0, maxChars);
      Diagnostic.trace(TRACE, "The whole fillbuffer looks like \"" + sb + "\"");
      Diagnostic.trace(TRACE, "Asking to append from pos " + i + " for " + iLen + " chars");
    }
*/    
    result.append(fillBuffer, i, iLen);
    Diagnostic.trace(TRACE, "returning " + result);
    return result;
  }
  
  /**
  This method will parse (or decompose) a string using the existing pattern.
  
  @param text This is the input String which is to be parsed.
  @param status Currently not used
  
  @return This method will allocate a new StringBuffer and fill it with the parsed version
  of the 'text' parameter.
  
*/
  
  public StringBuffer parse(String text, ParsePosition pos) {
    int maxChars = charCount + litCount;
    
    // Patternless strings parse merely by copying the input to the output
    if (maxChars == 0)
      return new StringBuffer(text);
    
    int i;
    StringBuffer result;
    int iLen = text.length();
    char buffer[] = new char[iLen];
    text.getChars(0, iLen, buffer, 0);
    int startPos;
    char[] fillBuffer = new char[maxChars];
    Diagnostic.trace(TRACE, "The text " + text + " is " + iLen + " and we will fill into " + fillBuffer.length);
    
    int nFilled = 0;
    int fillPos = 0;
    char c;
    if (!injectLiterals || litCount == 0) {
      Diagnostic.trace(TRACE, "No literal stripping -- removing up to " + iLen + " chars from " + text);
      for(i = 0; i < maxChars; ++i) {
        Diagnostic.trace(TRACE, "Buffer[" + i + "]");
        if (litMask != null && litMask[i] != NOT_A_CHAR && ((infoMask[i] & PASSWORD_CHAR) == 0)) {			// literals copy thru directly
          Diagnostic.trace(TRACE, " filling literal " + litMask[i]);
          //fillBuffer[fillPos++] = (litMask[i] == PasswordChar) ? '*' : litMask[i];
          fillBuffer[fillPos++] = litMask[i];
          continue;
        }
        c = (i < iLen)
            ? buffer[i]
            : fillChar;	// else comes form text -- beware running off end
        if (c == fillChar) {
          c = replaceChar;
          if (c == 0)					// also beware of fillChars turning into null
            continue;
        }
        Diagnostic.trace(TRACE, "  appending " + c);
        fillBuffer[fillPos++] = c;
      }
      Diagnostic.trace(TRACE, "And the final buffer has " + fillPos + " chars");
      result = new StringBuffer(fillPos);
      result.append(fillBuffer, 0, fillPos);
      return result;
    }
    
    i = 0;
    int incr = 1;
    int litPos = 0;
    int nRemaining = iLen;
    
    if (rightToLeft) {
      i = iLen - 1;
      incr = -1;
      fillPos = maxChars - 1;
      litPos = maxChars - 1;
    }
    
    for (; i >= 0 && i < iLen; i += incr, litPos += incr) {
      
      Diagnostic.trace(TRACE, "Buffer[" + i + "]");
      if (buffer[i] == fillChar) {		// fill chars are valid place holders but get stripped
        Diagnostic.trace(TRACE, "   fillchar -- skipping");
        continue;
      }
      if (litMask[litPos] == NOT_A_CHAR || ((infoMask[litPos] & PASSWORD_CHAR) != 0)) {
        c = buffer[i];
        Diagnostic.trace(TRACE, "   put '" + c + "' into fillBuffer[" + fillPos + "] and iLen = " + iLen);
        --nRemaining;
        if (c == fillChar) {
          Diagnostic.trace(TRACE, "     it is a fillChar and gets translated to" + replaceChar);
          c = replaceChar;
          if (c == 0) {
            Diagnostic.trace(TRACE, "      and just got stripped");
            continue;
          }
        }
        fillBuffer[fillPos] = c;
        fillPos += incr;
        ++nFilled;
        continue;
      }
      if (litMask[litPos] == buffer[i]) {
        Diagnostic.trace(TRACE, "  matched literal, skipping");
        continue;
      }
      //
      // Hmmm -- we are here if the character in text[i] doesn't match the literal we expected.
      // This means it was not a legal string w.r.t. this mask -- what to do... let's stop the loop
      break;
    }
    
    result = new StringBuffer(nFilled);
    if (rightToLeft)
      result.append(fillBuffer, maxChars-nFilled, nFilled);
    else
      result.append(fillBuffer, 0, nFilled);
    return result;
  }
  
  
  public final Object parseObject(String source, ParsePosition pos) {
    return parse(source, pos);
  }
  
  public final StringBuffer format(Object obj,
                                   StringBuffer toAppendTo,
                                   FieldPosition pos) {
    return format(obj.toString(), toAppendTo, pos);
  }

  private final boolean isLiteral(int position) {
    return (infoMask == null)
        ? false
        : ((position < 0) || (position >= infoMask.length))
          ? false
          : ((infoMask[position] & LITERAL_CHAR) != 0);
  }

  private final boolean isPassword(int position) {
    return (infoMask == null)
        ? false
        : ((position < 0) || (position >= infoMask.length))
          ? false
          : ((infoMask[position] & PASSWORD_CHAR) != 0);
  }
  
  private final boolean isOptional(int position) {
    return (infoMask == null)
        ? true
        : ((position < 0) || (position >= infoMask.length))
          ? true
          : ((infoMask[position] & OPTIONAL_CHAR) != 0);
  }
  
  private final boolean spotForCharacter(int position) {
    return (infoMask == null) ? true : ((infoMask[position] & LITERAL_CHAR) != 0);
/*    
    if (litMask == null)
      return true;
    if (litMask[position] == NOT_A_CHAR || ((infoMask[i] & PASSWORD_CHAR) != 0))
      return true;
    return false;
*/
  }
}
